import tempfile
import shutil
import os

import pytest

from errbot import repo_manager
from errbot.storage.memory import MemoryStoragePlugin

assets = os.path.join(os.path.dirname(__file__), 'assets')


@pytest.fixture
def plugdir_and_storage(request):
    plugins_dir = tempfile.mkdtemp()
    storage_plugin = MemoryStoragePlugin('repomgr')

    def on_finish():
        shutil.rmtree(plugins_dir)
    request.addfinalizer(on_finish)
    return plugins_dir, storage_plugin


def test_index_population(plugdir_and_storage):
    plugdir, storage = plugdir_and_storage
    manager = repo_manager.BotRepoManager(storage,
                                          plugdir,
                                          (os.path.join(assets, 'repos', 'simple.json'),))
    manager.index_update()

    index_entry = manager[repo_manager.REPO_INDEX]

    assert repo_manager.LAST_UPDATE in index_entry
    assert 'pluginname1' in index_entry['name1/err-reponame1']
    assert 'pluginname2' in index_entry['name2/err-reponame2']


def test_index_merge(plugdir_and_storage):
    plugdir, storage = plugdir_and_storage
    manager = repo_manager.BotRepoManager(storage,
                                          plugdir,
                                          (os.path.join(assets, 'repos', 'b.json'),
                                           os.path.join(assets, 'repos', 'a.json'),))
    manager.index_update()

    index_entry = manager[repo_manager.REPO_INDEX]

    # First they should be all here
    assert 'pluginname1' in index_entry['name1/err-reponame1']
    assert 'pluginname2' in index_entry['name2/err-reponame2']
    assert 'pluginname3' in index_entry['name3/err-reponame3']

    # then it must be the correct one of the overriden one

    assert index_entry['name2/err-reponame2']['pluginname2']['name'] == 'NewPluginName2'


def test_reverse_merge(plugdir_and_storage):
    plugdir, storage = plugdir_and_storage
    manager = repo_manager.BotRepoManager(storage,
                                          plugdir,
                                          (os.path.join(assets, 'repos', 'a.json'),
                                           os.path.join(assets, 'repos', 'b.json'),))
    manager.index_update()

    index_entry = manager[repo_manager.REPO_INDEX]
    assert not index_entry['name2/err-reponame2']['pluginname2']['name'] == 'NewPluginName2'


def test_no_update_if_one_fails(plugdir_and_storage):
    plugdir, storage = plugdir_and_storage
    manager = repo_manager.BotRepoManager(storage,
                                          plugdir,
                                          (os.path.join(assets, 'repos', 'a.json'),
                                           os.path.join(assets, 'repos', 'doh.json'),))
    manager.index_update()
    assert repo_manager.REPO_INDEX not in manager


def test_tokenization():
    e = {
        "python": "2+",
        "repo": "https://github.com/name/err-reponame1",
        "path": "/plugin1.plug",
        "avatar_url": "https://avatars.githubusercontent.com/u/588833?v=3",
        "name": "PluginName1",
        "documentation": "docs1"
    }
    words = {
        'https',
        'com',
        'name',
        'err',
        'docs1',
        'reponame1',
        'plug',
        '2',
        'plugin1',
        'avatars',
        'github',
        'githubusercontent',
        'u',
        'v',
        '3',
        '588833',
        'pluginname1'
    }
    assert repo_manager.tokenizeJsonEntry(e) == words


def test_search(plugdir_and_storage):
    plugdir, storage = plugdir_and_storage
    manager = repo_manager.BotRepoManager(storage,
                                          plugdir,
                                          (os.path.join(assets, 'repos', 'simple.json'),))

    a = [p for p in manager.search_repos('docs2')]
    assert len(a) == 1
    assert a[0].name == 'pluginname2'

    a = [p for p in manager.search_repos('zorg')]
    assert len(a) == 0

    a = [p for p in manager.search_repos('plug')]
    assert len(a) == 2


def test_git_url_name_guessing():
    assert repo_manager.human_name_for_git_url('https://github.com/errbotio/err-imagebot.git') \
        == 'errbotio/err-imagebot'



from os import path

# This is to test end2end i18n behavior.

extra_plugin_dir = path.join(path.dirname(path.realpath(__file__)), 'template_plugin')


def test_templates_1(testbot):
    assert 'ok' in testbot.exec_command('!test template1')


def test_templates_2(testbot):
    assert 'ok' in testbot.exec_command('!test template2')


def test_templates_3(testbot):
    assert 'ok' in testbot.exec_command('!test template3')


def test_templates_4(testbot):
    assert 'ok' in testbot.exec_command('!test template4 ok')


def test_templates_5(testbot):
    assert 'the following arguments are required: my_var' in testbot.exec_command('!test template4')



from os import path

extra_plugin_dir = path.join(path.dirname(path.realpath(__file__)), 'fail_config_plugin')


def test_failed_config(testbot):
    assert 'Incorrect plugin configuration: Message explaining why it failed.' \
           in testbot.exec_command('!plugin config Failp {}')



import logging
import os
import re
import sys
from jinja2 import Environment, FileSystemLoader

import errbot
from errbot.backends.base import Message, ONLINE
from errbot.backends.text import TextBackend   # we use that as we emulate MUC there already
from errbot.rendering import xhtml

CARD_TMPL = Environment(loader=FileSystemLoader(os.path.dirname(__file__)),
                        autoescape=True).get_template('graphic_card.html')

log = logging.getLogger(__name__)

try:
    from PySide import QtCore, QtGui, QtWebKit
    from PySide.QtGui import QCompleter
    from PySide.QtCore import Qt
except ImportError:
    log.exception("Could not start the graphical backend")
    log.fatal(""" To install graphic support use:
    pip install errbot[graphic]
    """)
    sys.exit(-1)


class CommandBox(QtGui.QPlainTextEdit, object):
    newCommand = QtCore.Signal(str)

    def reset_history(self):
        self.history_index = len(self.history)

    def __init__(self, history, commands, prefix):
        self.history_index = 0
        self.history = history
        self.reset_history()
        self.prefix = prefix
        super().__init__()

        # Autocompleter
        self.completer = None
        self.updateCompletion(commands)
        self.autocompleteStart = None

    def updateCompletion(self, commands):
        if self.completer:
            self.completer.activated.disconnect(self.onAutoComplete)
        self.completer = QCompleter([(self.prefix + name).replace('_', ' ', 1) for name in commands], self)
        self.completer.setCaseSensitivity(Qt.CaseInsensitive)
        self.completer.setWidget(self)
        self.completer.activated.connect(self.onAutoComplete)

    def onAutoComplete(self, text):
        # Select the text from autocompleteStart until the current cursor
        cursor = self.textCursor()
        cursor.setPosition(0, cursor.KeepAnchor)
        # Replace it with the selected text
        cursor.insertText(text)
        self.autocompleteStart = None

    # noinspection PyStringFormat
    def keyPressEvent(self, *args, **kwargs):
        event = args[0]
        key = event.key()
        ctrl = event.modifiers() == QtCore.Qt.ControlModifier
        alt = event.modifiers() == QtCore.Qt.AltModifier

        # don't disturb the completer behavior
        if self.completer.popup().isVisible() and key in (Qt.Key_Enter, Qt.Key_Return, Qt.Key_Tab, Qt.Key_Backtab):
            event.ignore()
            return

        if self.autocompleteStart is not None and not event.text().isalnum() and \
                not (key == Qt.Key_Backspace and self.textCursor().position() > self.autocompleteStart):
            self.completer.popup().hide()
            self.autocompleteStart = None

        if key == Qt.Key_Space and (ctrl or alt):
            # Pop-up the autocompleteList
            rect = self.cursorRect(self.textCursor())
            rect.setSize(QtCore.QSize(300, 500))
            self.autocompleteStart = self.textCursor().position()
            self.completer.complete(rect)  # The popup is positioned in the next if block

        if self.autocompleteStart:
            prefix = self.toPlainText()
            cur = self.textCursor()
            cur.setPosition(self.autocompleteStart)

            self.completer.setCompletionPrefix(prefix)
            # Select the first one of the matches
            self.completer.popup().setCurrentIndex(self.completer.completionModel().index(0, 0))

        if key == Qt.Key_Up:
            if self.history_index > 0:
                self.history_index -= 1
                self.setPlainText(f'{self.prefix}{" ".join(self.history[self.history_index])}')
                return
        elif key == Qt.Key_Down:
            if self.history_index < len(self.history) - 1:
                self.history_index += 1
                self.setPlainText(f'{self.prefix}{" ".join(self.history[self.history_index])}')
                return
        elif key == QtCore.Qt.Key_Return and (ctrl or alt):
            self.newCommand.emit(self.toPlainText())
            self.reset_history()
        super().keyPressEvent(*args, **kwargs)


urlfinder = re.compile(r'http([^.\s]+\.[^.\s]*)+[^.\s]{2,}')

backends_path = os.path.join(os.path.dirname(errbot.__file__), 'backends')

images_path = os.path.join(backends_path, 'images')
prompt_path = os.path.join(images_path, 'prompt.svg')
icon_path = os.path.join(images_path, 'errbot.svg')
bg_path = os.path.join(images_path, 'errbot-bg.svg')

style_path = os.path.join(backends_path, 'styles')
css_path = os.path.join(style_path, 'style.css')
demo_css_path = os.path.join(style_path, 'style-demo.css')

TOP = f'<html><body style="background-image: url(\'file://{bg_path}\');">'
BOTTOM = '</body></html>'


class ChatApplication(QtGui.QApplication):
    newAnswer = QtCore.Signal(str)

    def __init__(self, bot):
        self.bot = bot
        super().__init__(sys.argv)
        self.mainW = QtGui.QWidget()
        self.mainW.setWindowTitle('Errbot')
        self.mainW.setWindowIcon(QtGui.QIcon(icon_path))
        vbox = QtGui.QVBoxLayout()
        help_label = QtGui.QLabel("ctrl or alt+space for autocomplete -- ctrl or alt+Enter to send your message")
        self.input = CommandBox(bot.cmd_history[str(bot.user)], bot.all_commands, bot.bot_config.BOT_PREFIX)
        self.demo_mode = hasattr(bot.bot_config, 'TEXT_DEMO_MODE') and bot.bot_config.TEXT_DEMO_MODE
        font = QtGui.QFont("Arial", QtGui.QFont.Bold)
        font.setPointSize(30 if self.demo_mode else 15)
        self.input.setFont(font)

        self.output = QtWebKit.QWebView()
        css = demo_css_path if self.demo_mode else css_path
        self.output.settings().setUserStyleSheetUrl(QtCore.QUrl.fromLocalFile(css))

        # init webpage
        self.buffer = ""
        self.update_webpage()

        # layout
        vbox.addWidget(self.output)
        vbox.addWidget(self.input)
        vbox.addWidget(help_label)
        self.mainW.setLayout(vbox)

        # setup web view to open liks in external browser
        self.output.page().setLinkDelegationPolicy(QtWebKit.QWebPage.DelegateAllLinks)

        # connect signals/slots
        self.output.page().mainFrame().contentsSizeChanged.connect(self.scroll_output_to_bottom)
        self.output.page().linkClicked.connect(QtGui.QDesktopServices.openUrl)
        self.input.newCommand.connect(lambda text: bot.send_command(text))
        self.newAnswer.connect(self.new_message)
        if self.demo_mode:
            self.mainW.showFullScreen()
        else:
            self.mainW.show()

    def new_message(self, text, receiving=True):
        size = 50 if self.demo_mode else 25
        user = f'<img src="file://{prompt_path}" height={size:d} />'
        bot = f'<img src="file://{icon_path}" height={size:d}/>'
        self.buffer += f'<div class="{"receiving" if receiving else "sending"}">{bot if receiving else user}' \
                       f'<br/>{text}</div>'

        self.update_webpage()

    def update_webpage(self):
        self.output.setHtml(TOP + self.buffer + BOTTOM)

    def scroll_output_to_bottom(self):
        self.output.page().mainFrame().scroll(0, self.output.page().mainFrame().scrollBarMaximum(QtCore.Qt.Vertical))

    def update_commands(self, commands):
        self.input.updateCompletion(commands)


class GraphicBackend(TextBackend):
    def __init__(self, config):
        super().__init__(config)
        # create window and components
        self.md = xhtml()
        self.app = ChatApplication(self)

    def connect_callback(self):
        super().connect_callback()
        self.app.update_commands(self.all_commands)

    def send_command(self, text):
        self.app.new_message(text, False)
        msg = Message(text)
        msg.frm = self.user
        msg.to = self.bot_identifier  # To me only
        self.callback_message(msg)
        # implements the mentions.
        mentioned = [self.build_identifier(word[1:]) for word in re.findall(r"@[\w']+", text)
                     if word.startswith('@')]
        if mentioned:
            self.callback_mention(msg, mentioned)

        self.app.input.clear()

    def build_message(self, text):
        msg = Message(text)
        msg.frm = self.bot_identifier
        return msg  # rebuild a pure html snippet to include directly in the console html

    def send_message(self, msg):
        if hasattr(msg, 'body') and msg.body and not msg.body.isspace():
            content = self.md.convert(msg.body)
            log.debug("html:\n%s", content)
            self.app.newAnswer.emit(content)

    def send_card(self, card):
        self.app.newAnswer.emit(CARD_TMPL.render(card=card))

    def change_presence(self, status: str = ONLINE, message: str = '') -> None:
        pass

    def serve_forever(self):
        self.connect_callback()  # notify that the connection occured

        try:
            self.app.exec_()
        finally:
            self.disconnect_callback()
            self.shutdown()
            exit(0)

    @property
    def mode(self):
        return 'graphic'

    def prefix_groupchat_reply(self, message, identifier):
        super().prefix_groupchat_reply(message, identifier)
        message.body = f'@{identifier.nick} {message.body}'



import logging
import sys
from functools import lru_cache

from threading import Thread
from time import sleep

from errbot.backends.base import Message, Room, Presence, RoomNotJoinedError, Identifier, RoomOccupant, Person
from errbot.backends.base import ONLINE, OFFLINE, AWAY, DND
from errbot.core import ErrBot
from errbot.rendering import text, xhtml, xhtmlim

log = logging.getLogger(__name__)

try:
    from sleekxmpp import ClientXMPP
    from sleekxmpp.xmlstream import resolver, cert
    from sleekxmpp import JID
    from sleekxmpp.exceptions import IqError

except ImportError:
    log.exception("Could not start the XMPP backend")
    log.fatal("""
    If you intend to use the XMPP backend please install the support for XMPP with:
    pip install errbot[XMPP]
    """)
    sys.exit(-1)

# LRU to cache the JID queries.
IDENTIFIERS_LRU = 1024


class XMPPIdentifier(Identifier):
    """
    This class is the parent and the basic contract of all the ways the backends
    are identifying a person on their system.
    """

    def __init__(self, node, domain, resource):
        if not node:
            raise Exception('An XMPPIdentifier needs to have a node.')
        if not domain:
            raise Exception('An XMPPIdentifier needs to have a domain.')
        self._node = node
        self._domain = domain
        self._resource = resource

    @property
    def node(self):
        return self._node

    @property
    def domain(self):
        return self._domain

    @property
    def resource(self):
        return self._resource

    @property
    def person(self):
        return self._node + '@' + self._domain

    @property
    def nick(self):
        return self._node

    @property
    def fullname(self):
        return None  # Not supported by default on XMPP.

    @property
    def client(self):
        return self._resource

    def __str__(self):
        answer = self._node + '@' + self._domain  # don't call .person: see below
        if self._resource:
            answer += '/' + self._resource
        return answer

    def __unicode__(self):
        return str(self.__str__())

    def __eq__(self, other):
        if not isinstance(other, XMPPIdentifier):
            log.debug("Weird, you are comparing an XMPPIdentifier to a %s", type(other))
            return False
        return self._domain == other._domain and self._node == other._node and self._resource == other._resource


class XMPPPerson(XMPPIdentifier, Person):
    aclattr = XMPPIdentifier.person

    def __eq__(self, other):
        if not isinstance(other, XMPPPerson):
            log.debug("Weird, you are comparing an XMPPPerson to a %s", type(other))
            return False
        return self._domain == other._domain and self._node == other._node


class XMPPRoom(XMPPIdentifier, Room):
    def __init__(self, room_jid, bot):
        self._bot = bot
        self.xep0045 = self._bot.conn.client.plugin['xep_0045']
        node, domain, resource = split_identifier(room_jid)
        super().__init__(node, domain, resource)

    def join(self, username=None, password=None):
        """
        Join the room.

        If the room does not exist yet, this will automatically call
        :meth:`create` on it first.
        """
        room = str(self)
        self.xep0045.joinMUC(room, username, password=password, wait=True)
        self._bot.conn.add_event_handler(f'muc::{room}::got_online', self._bot.user_joined_chat)
        self._bot.conn.add_event_handler(f'muc::{room}::got_offline', self._bot.user_left_chat)
        # Room configuration can only be done once a MUC presence stanza
        # has been received from the server. This HAS to take place in a
        # separate thread because of how SleekXMPP processes these stanzas.
        t = Thread(target=self.configure)
        t.setDaemon(True)
        t.start()
        self._bot.callback_room_joined(self)
        log.info('Joined room %s.', room)

    def leave(self, reason=None):
        """
        Leave the room.

        :param reason:
            An optional string explaining the reason for leaving the room
        """
        if reason is None:
            reason = ""
        room = str(self)
        try:
            self.xep0045.leaveMUC(room=room, nick=self.xep0045.ourNicks[room], msg=reason)

            self._bot.conn.del_event_handler(f'muc::{room}::got_online', self._bot.user_joined_chat)
            self._bot.conn.del_event_handler(f'muc::{room}::got_offline', self._bot.user_left_chat)
            log.info('Left room %s.', room)
            self._bot.callback_room_left(self)
        except KeyError:
            log.debug('Trying to leave %s while not in this room.', room)

    def create(self):
        """
        Not supported on this back-end (SleekXMPP doesn't support it).
        Will join the room to ensure it exists, instead.
        """
        logging.warning(
            "XMPP back-end does not support explicit creation, joining room "
            "instead to ensure it exists."
        )
        self.join(username=str(self))

    def destroy(self):
        """
        Destroy the room.

        Calling this on a non-existing room is a no-op.
        """
        self.xep0045.destroy(str(self))
        log.info('Destroyed room %s.', self)

    @property
    def exists(self):
        """
        Boolean indicating whether this room already exists or not.

        :getter:
            Returns `True` if the room exists, `False` otherwise.
        """
        logging.warning(
            'XMPP back-end does not support determining if a room exists. Returning the result of joined instead.')
        return self.joined

    @property
    def joined(self):
        """
        Boolean indicating whether this room has already been joined.

        :getter:
            Returns `True` if the room has been joined, `False` otherwise.
        """
        return str(self) in self.xep0045.getJoinedRooms()

    @property
    def topic(self):
        """
        The room topic.

        :getter:
            Returns the topic (a string) if one is set, `None` if no
            topic has been set at all.
        :raises:
            :class:`~RoomNotJoinedError` if the room has not yet been joined.
        """
        if not self.joined:
            raise RoomNotJoinedError("Must be in a room in order to see the topic.")
        try:
            return self._bot._room_topics[str(self)]
        except KeyError:
            return None

    @topic.setter
    def topic(self, topic):
        """
        Set the room's topic.

        :param topic:
            The topic to set.
        """
        # Not supported by SleekXMPP at the moment :(
        raise NotImplementedError("Setting the topic is not supported on this back-end.")

    @property
    def occupants(self):
        """
        The room's occupants.

        :getter:
            Returns a list of :class:`~errbot.backends.base.MUCOccupant` instances.
        :raises:
            :class:`~MUCNotJoinedError` if the room has not yet been joined.
        """
        occupants = []
        try:
            for occupant in self.xep0045.rooms[str(self)].values():
                room_node, room_domain, _ = split_identifier(occupant['room'])
                nick = occupant['nick']
                occupants.append(XMPPRoomOccupant(room_node, room_domain, nick, self))
        except KeyError:
            raise RoomNotJoinedError("Must be in a room in order to see occupants.")
        return occupants

    def invite(self, *args):
        """
        Invite one or more people into the room.

        :*args:
            One or more JID's to invite into the room.
        """
        room = str(self)
        for jid in args:
            self.xep0045.invite(room, jid)
            log.info('Invited %s to %s.', jid, room)

    def configure(self):
        """
        Configure the room.

        Currently this simply sets the default room configuration as
        received by the server. May be extended in the future to set
        a custom room configuration instead.
        """
        room = str(self)
        affiliation = None
        while affiliation is None:
            sleep(0.5)
            affiliation = self.xep0045.getJidProperty(
                room=room,
                nick=self.xep0045.ourNicks[room],
                jidProperty='affiliation'
            )

        if affiliation == "owner":
            log.debug('Configuring room %s: we have owner affiliation.', room)
            form = self.xep0045.getRoomConfig(room)
            self.xep0045.configureRoom(room, form)
        else:
            log.debug("Not configuring room %s: we don't have owner affiliation (affiliation=%s)", room, affiliation)


class XMPPRoomOccupant(XMPPPerson, RoomOccupant):
    def __init__(self, node, domain, resource, room):
        super().__init__(node, domain, resource)
        self._room = room

    @property
    def person(self):
        return str(self)  # this is the full identifier.

    @property
    def real_jid(self):
        """
        The JID of the room occupant, they used to login.
        Will only work if the errbot is moderator in the MUC or it is not anonymous.
        """
        room_jid = self._node + '@' + self._domain
        jid = JID(self._room.xep0045.getJidProperty(room_jid, self.resource, 'jid'))
        return jid.bare

    @property
    def room(self):
        return self._room

    nick = XMPPPerson.resource


class XMPPConnection(object):
    def __init__(self, jid, password, feature=None, keepalive=None,
                 ca_cert=None, server=None, use_ipv6=None, bot=None):
        if feature is None:
            feature = {}
        self._bot = bot
        self.connected = False
        self.server = server
        self.client = ClientXMPP(jid, password, plugin_config={'feature_mechanisms': feature})
        self.client.register_plugin('xep_0030')  # Service Discovery
        self.client.register_plugin('xep_0045')  # Multi-User Chat
        self.client.register_plugin('xep_0199')  # XMPP Ping
        self.client.register_plugin('xep_0203')  # XMPP Delayed messages
        self.client.register_plugin('xep_0249')  # XMPP direct MUC invites

        if keepalive is not None:
            self.client.whitespace_keepalive = True  # Just in case SleekXMPP's default changes to False in the future
            self.client.whitespace_keepalive_interval = keepalive

        if use_ipv6 is not None:
            self.client.use_ipv6 = use_ipv6

        self.client.ca_certs = ca_cert  # Used for TLS certificate validation

        self.client.add_event_handler("session_start", self.session_start)

    def session_start(self, _):
        self.client.send_presence()
        self.client.get_roster()

    def connect(self):
        if not self.connected:
            if self.server is not None:
                self.client.connect(self.server)
            else:
                self.client.connect()
            self.connected = True
        return self

    def disconnect(self):
        self.client.disconnect(wait=True)
        self.connected = False

    def serve_forever(self):
        self.client.process(block=True)

    def add_event_handler(self, name, cb):
        self.client.add_event_handler(name, cb)

    def del_event_handler(self, name, cb):
        self.client.del_event_handler(name, cb)


XMPP_TO_ERR_STATUS = {'available': ONLINE,
                      'away': AWAY,
                      'dnd': DND,
                      'unavailable': OFFLINE}


def split_identifier(txtrep):
    split_jid = txtrep.split('@', 1)
    node, domain = '@'.join(split_jid[:-1]), split_jid[-1]
    if domain.find('/') != -1:
        domain, resource = domain.split('/', 1)
    else:
        resource = None

    return node, domain, resource


class XMPPBackend(ErrBot):
    room_factory = XMPPRoom
    roomoccupant_factory = XMPPRoomOccupant

    def __init__(self, config):
        super().__init__(config)
        identity = config.BOT_IDENTITY

        self.jid = identity['username']  # backward compatibility
        self.password = identity['password']
        self.server = identity.get('server', None)
        self.feature = config.__dict__.get('XMPP_FEATURE_MECHANISMS', {})
        self.keepalive = config.__dict__.get('XMPP_KEEPALIVE_INTERVAL', None)
        self.ca_cert = config.__dict__.get('XMPP_CA_CERT_FILE', '/etc/ssl/certs/ca-certificates.crt')
        self.xhtmlim = config.__dict__.get('XMPP_XHTML_IM', False)
        self.use_ipv6 = config.__dict__.get('XMPP_USE_IPV6', None)

        # generic backend compatibility
        self.bot_identifier = self._build_person(self.jid)

        self.conn = self.create_connection()
        self.conn.add_event_handler("message", self.incoming_message)
        self.conn.add_event_handler("session_start", self.connected)
        self.conn.add_event_handler("disconnected", self.disconnected)
        # presence related handlers
        self.conn.add_event_handler("got_online", self.contact_online)
        self.conn.add_event_handler("got_offline", self.contact_offline)
        self.conn.add_event_handler("changed_status", self.user_changed_status)
        # MUC subject events
        self.conn.add_event_handler("groupchat_subject", self.chat_topic)
        self._room_topics = {}
        self.md_xhtml = xhtml()
        self.md_text = text()

    def create_connection(self):
        return XMPPConnection(
            jid=self.jid,  # textual and original representation
            password=self.password,
            feature=self.feature,
            keepalive=self.keepalive,
            ca_cert=self.ca_cert,
            server=self.server,
            use_ipv6=self.use_ipv6,
            bot=self
        )

    def _build_room_occupant(self, txtrep):
        node, domain, resource = split_identifier(txtrep)
        return self.roomoccupant_factory(node, domain, resource, self.query_room(node + '@' + domain))

    def _build_person(self, txtrep):
        return XMPPPerson(*split_identifier(txtrep))

    def incoming_message(self, xmppmsg):
        """Callback for message events"""
        if xmppmsg['type'] == "error":
            log.warning("Received error message: %s", xmppmsg)
            return

        msg = Message(xmppmsg['body'])
        if 'html' in xmppmsg.keys():
            msg.html = xmppmsg['html']
        log.debug("incoming_message from: %s", msg.frm)
        if xmppmsg['type'] == 'groupchat':
            msg.frm = self._build_room_occupant(xmppmsg['from'].full)
            msg.to = msg.frm.room
        else:
            msg.frm = self._build_person(xmppmsg['from'].full)
            msg.to = self._build_person(xmppmsg['to'].full)

        msg.nick = xmppmsg['mucnick']
        msg.delayed = bool(xmppmsg['delay']._get_attr('stamp'))  # this is a bug in sleekxmpp it should be ['from']
        self.callback_message(msg)

    def _idd_from_event(self, event):
        txtrep = event['from'].full
        return self._build_room_occupant(txtrep) if 'muc' in event else self._build_person(txtrep)

    def contact_online(self, event):
        log.debug('contact_online %s.', event)
        self.callback_presence(Presence(identifier=self._idd_from_event(event), status=ONLINE))

    def contact_offline(self, event):
        log.debug('contact_offline %s.', event)
        self.callback_presence(Presence(identifier=self._idd_from_event(event), status=OFFLINE))

    def user_joined_chat(self, event):
        log.debug('user_join_chat %s', event)
        self.callback_presence(Presence(identifier=self._idd_from_event(event), status=ONLINE))

    def user_left_chat(self, event):
        log.debug('user_left_chat %s', event)
        self.callback_presence(Presence(identifier=self._idd_from_event(event), status=OFFLINE))

    def chat_topic(self, event):
        log.debug("chat_topic %s.", event)
        room = event.values['mucroom']
        topic = event.values['subject']
        if topic == "":
            topic = None
        self._room_topics[room] = topic
        room = XMPPRoom(event.values['mucroom'], self)
        self.callback_room_topic(room)

    def user_changed_status(self, event):
        log.debug('user_changed_status %s.', event)
        errstatus = XMPP_TO_ERR_STATUS.get(event['type'], None)
        message = event['status']
        if not errstatus:
            errstatus = event['type']
        self.callback_presence(Presence(identifier=self._idd_from_event(event), status=errstatus, message=message))

    def connected(self, data):
        """Callback for connection events"""
        self.connect_callback()

    def disconnected(self, data):
        """Callback for disconnection events"""
        self.disconnect_callback()

    def send_message(self, msg):
        super().send_message(msg)

        log.debug('send_message to %s', msg.to)

        # We need to unescape the unicode characters (not the markup incompatible ones)
        mhtml = xhtmlim.unescape(self.md_xhtml.convert(msg.body)) if self.xhtmlim else None

        self.conn.client.send_message(mto=str(msg.to),
                                      mbody=self.md_text.convert(msg.body),
                                      mhtml=mhtml,
                                      mtype='chat' if msg.is_direct else 'groupchat')

    def change_presence(self, status: str = ONLINE, message: str = '') -> None:
        log.debug('Change bot status to %s, message %s.', status, message)
        self.conn.client.send_presence(pshow=status, pstatus=message)

    def serve_forever(self):
        self.conn.connect()

        try:
            self.conn.serve_forever()
        finally:
            log.debug("Trigger disconnect callback")
            self.disconnect_callback()
            log.debug("Trigger shutdown")
            self.shutdown()

    @lru_cache(IDENTIFIERS_LRU)
    def build_identifier(self, txtrep):
        log.debug('build identifier for %s', txtrep)
        try:
            xep0030 = self.conn.client.plugin['xep_0030']
            info = xep0030.get_info(jid=txtrep)
            disco_info = info['disco_info']
            if disco_info:  # Hipchat can return an empty response here.
                for category, typ, _, name in disco_info['identities']:
                    if category == 'conference':
                        log.debug('This is a room ! %s', txtrep)
                        return self.query_room(txtrep)
                    if category == 'client' and 'http://jabber.org/protocol/muc' in info['disco_info']['features']:
                        log.debug('This is room occupant ! %s', txtrep)
                        return self._build_room_occupant(txtrep)
        except IqError as iq:
            log.debug('xep_0030 is probably not implemented on this server. %s.', iq)
        log.debug('This is a person ! %s', txtrep)
        return self._build_person(txtrep)

    def build_reply(self, msg, text=None, private=False, threaded=False):
        response = self.build_message(text)
        response.frm = self.bot_identifier

        if msg.is_group and not private:
            # stripped returns the full bot@conference.domain.tld/chat_username
            # but in case of a groupchat, we should only try to send to the MUC address
            # itself (bot@conference.domain.tld)
            response.to = XMPPRoom(msg.frm.node + '@' + msg.frm.domain, self)
        elif msg.is_direct:
            # preserve from in case of a simple chat message.
            # it is either a user to user or user_in_chatroom to user case.
            # so we need resource.
            response.to = msg.frm
        elif hasattr(msg.to, 'person') and msg.to.person == self.bot_config.BOT_IDENTITY['username']:
            # This is a direct private message, not initiated through a MUC. Use
            # stripped to remove the resource so that the response goes to the
            # client with the highest priority
            response.to = XMPPPerson(msg.frm.node, msg.frm.domain, None)
        else:
            # This is a private message that was initiated through a MUC. Don't use
            # stripped here to retain the resource, else the XMPP server doesn't
            # know which user we're actually responding to.
            response.to = msg.frm
        return response

    @property
    def mode(self):
        return 'xmpp'

    def rooms(self):
        """
        Return a list of rooms the bot is currently in.

        :returns:
            A list of :class:`~errbot.backends.base.XMPPMUCRoom` instances.
        """
        xep0045 = self.conn.client.plugin['xep_0045']
        return [XMPPRoom(room, self) for room in xep0045.getJoinedRooms()]

    def query_room(self, room):
        """
        Query a room for information.

        :param room:
            The JID/identifier of the room to query for.
        :returns:
            An instance of :class:`~XMPPMUCRoom`.
        """
        return XMPPRoom(room, self)

    def prefix_groupchat_reply(self, message, identifier):
        super().prefix_groupchat_reply(message, identifier)
        message.body = f'@{identifier.nick} {message.body}'

    def __hash__(self):
        return 0



import logging
import sys

from errbot.backends.base import RoomError, Identifier, Person, RoomOccupant, Stream, ONLINE, Room
from errbot.core import ErrBot
from errbot.rendering import text
from errbot.rendering.ansiext import enable_format, TEXT_CHRS

log = logging.getLogger(__name__)

TELEGRAM_MESSAGE_SIZE_LIMIT = 1024
UPDATES_OFFSET_KEY = '_telegram_updates_offset'

try:
    import telegram
except ImportError:
    log.exception("Could not start the Telegram back-end")
    log.fatal(
        "You need to install the telegram support in order "
        "to use the Telegram backend.\n"
        "You should be able to install this package using:\n"
        "pip install errbot[telegram]"
    )
    sys.exit(1)


class RoomsNotSupportedError(RoomError):
    def __init__(self, message=None):
        if message is None:
            message = (
                "Room operations are not supported on Telegram. "
                "While Telegram itself has groupchat functionality, it does not "
                "expose any APIs to bots to get group membership or otherwise "
                "interact with groupchats."
            )
        super().__init__(message)


class TelegramBotFilter(object):
    """
    This is a filter for the logging library that filters the
    "No new updates found." log message generated by telegram.bot.

    This is an INFO-level log message that gets logged for every
    getUpdates() call where there are no new messages, so is way
    too verbose.
    """

    @staticmethod
    def filter(record):
        if record.getMessage() == "No new updates found.":
            return 0


class TelegramIdentifier(Identifier):
    def __init__(self, id):
        self._id = str(id)

    @property
    def id(self):
        return self._id

    def __unicode__(self):
        return str(self._id)

    def __eq__(self, other):
        return self._id == other.id

    __str__ = __unicode__

    aclattr = id


class TelegramPerson(TelegramIdentifier, Person):
    def __init__(self, id, first_name=None, last_name=None, username=None):
        super().__init__(id)
        self._first_name = first_name
        self._last_name = last_name
        self._username = username

    @property
    def id(self):
        return self._id

    @property
    def first_name(self):
        return self._first_name

    @property
    def last_name(self):
        return self._last_name

    @property
    def fullname(self):
        fullname = self.first_name
        if self.last_name is not None:
            fullname += " " + self.last_name
        return fullname

    @property
    def username(self):
        return self._username

    @property
    def client(self):
        return None

    person = id
    nick = username


class TelegramRoom(TelegramIdentifier, Room):
    def __init__(self, id, title=None):
        super().__init__(id)
        self._title = title

    @property
    def id(self):
        return self._id

    @property
    def title(self):
        """Return the groupchat title (only applies to groupchats)"""
        return self._title

    def join(self, username: str = None, password: str = None):
        raise RoomsNotSupportedError()

    def create(self):
        raise RoomsNotSupportedError()

    def leave(self, reason: str = None):
        raise RoomsNotSupportedError()

    def destroy(self):
        raise RoomsNotSupportedError()

    @property
    def joined(self):
        raise RoomsNotSupportedError()

    @property
    def exists(self):
        raise RoomsNotSupportedError()

    @property
    def topic(self):
        raise RoomsNotSupportedError()

    @property
    def occupants(self):
        raise RoomsNotSupportedError()

    def invite(self, *args):
        raise RoomsNotSupportedError()


class TelegramMUCOccupant(TelegramPerson, RoomOccupant):
    """
    This class represents a person inside a MUC.
    """

    def __init__(self, id, room, first_name=None, last_name=None, username=None):
        super().__init__(id=id, first_name=first_name, last_name=last_name, username=username)
        self._room = room

    @property
    def room(self):
        return self._room

    @property
    def username(self):
        return self._username


class TelegramBackend(ErrBot):
    def __init__(self, config):
        super().__init__(config)
        config.MESSAGE_SIZE_LIMIT = TELEGRAM_MESSAGE_SIZE_LIMIT
        logging.getLogger('telegram.bot').addFilter(TelegramBotFilter())

        identity = config.BOT_IDENTITY
        self.token = identity.get('token', None)
        if not self.token:
            log.fatal(
                "You need to supply a token for me to use. You can obtain "
                "a token by registering your bot with the Bot Father (@BotFather)"
            )
            sys.exit(1)
        self.telegram = None  # Will be initialized in serve_once
        self.bot_instance = None  # Will be set in serve_once

        compact = config.COMPACT_OUTPUT if hasattr(config, 'COMPACT_OUTPUT') else False
        enable_format('text', TEXT_CHRS, borders=not compact)
        self.md_converter = text()

    def serve_once(self):
        log.info("Initializing connection")
        try:
            self.telegram = telegram.Bot(token=self.token)
            me = self.telegram.getMe()
        except telegram.TelegramError as e:
            log.error("Connection failure: %s", e.message)
            return False

        self.bot_identifier = TelegramPerson(
            id=me.id,
            first_name=me.first_name,
            last_name=me.last_name,
            username=me.username
        )

        log.info("Connected")
        self.reset_reconnection_count()
        self.connect_callback()

        try:
            offset = self[UPDATES_OFFSET_KEY]
        except KeyError:
            offset = 0

        try:
            while True:
                log.debug("Getting updates with offset %s", offset)
                for update in self.telegram.getUpdates(offset=offset, timeout=60):
                    offset = update.update_id + 1
                    self[UPDATES_OFFSET_KEY] = offset
                    log.debug("Processing update: %s", update)
                    if not hasattr(update, 'message'):
                        log.warning("Unknown update type (no message present)")
                        continue
                    try:
                        self._handle_message(update.message)
                    except Exception:
                        log.exception("An exception occurred while processing update")
                log.debug("All updates processed, new offset is %s", offset)
        except KeyboardInterrupt:
            log.info("Interrupt received, shutting down..")
            return True
        except Exception:
            log.exception("Error reading from Telegram updates stream:")
        finally:
            log.debug("Triggering disconnect callback")
            self.disconnect_callback()

    def _handle_message(self, message):
        """
        Handle a received message.

        :param message:
            A message with a structure as defined at
            https://core.telegram.org/bots/api#message
        """
        if message.text is None:
            log.warning("Unhandled message type (not a text message) ignored")
            return

        message_instance = self.build_message(message.text)
        if message.chat['type'] == 'private':
            message_instance.frm = TelegramPerson(
                id=message.from_user.id,
                first_name=message.from_user.first_name,
                last_name=message.from_user.last_name,
                username=message.from_user.username
            )
            message_instance.to = self.bot_identifier
        else:
            room = TelegramRoom(id=message.chat.id, title=message.chat.title)
            message_instance.frm = TelegramMUCOccupant(
                id=message.from_user.id,
                room=room,
                first_name=message.from_user.first_name,
                last_name=message.from_user.last_name,
                username=message.from_user.username
            )
            message_instance.to = room
        message_instance.extras['message_id'] = message.message_id
        self.callback_message(message_instance)

    def send_message(self, msg):
        super().send_message(msg)
        body = self.md_converter.convert(msg.body)
        try:
            self.telegram.sendMessage(msg.to.id, body)
        except Exception:
            log.exception(
                f'An exception occurred while trying to send the following message to {msg.to.id}: {msg.body}')
            raise

    def change_presence(self, status: str = ONLINE, message: str = '') -> None:
        # It looks like telegram doesn't supports online presence for privacy reason.
        pass

    def build_identifier(self, txtrep):
        """
        Convert a textual representation into a :class:`~TelegramPerson` or :class:`~TelegramRoom`.
        """
        log.debug('building an identifier from %s.', txtrep)
        if not self._is_numeric(txtrep):
            raise ValueError('Telegram identifiers must be numeric.')
        id_ = int(txtrep)
        if id_ > 0:
            return TelegramPerson(id=id_)
        else:
            return TelegramRoom(id=id_)

    def build_reply(self, msg, text=None, private=False, threaded=False):
        response = self.build_message(text)
        response.frm = self.bot_identifier
        if private:
            response.to = msg.frm
        else:
            response.to = msg.frm if msg.is_direct else msg.to
        return response

    @property
    def mode(self):
        return 'telegram'

    def query_room(self, room):
        """
        Not supported on Telegram.

        :raises: :class:`~RoomsNotSupportedError`
        """
        raise RoomsNotSupportedError()

    def rooms(self):
        """
        Not supported on Telegram.

        :raises: :class:`~RoomsNotSupportedError`
        """
        raise RoomsNotSupportedError()

    def prefix_groupchat_reply(self, message, identifier):
        super().prefix_groupchat_reply(message, identifier)
        message.body = f'@{identifier.nick}: {message.body}'

    def _telegram_special_message(self, chat_id, content, msg_type, **kwargs):
        """Send special message."""
        if msg_type == 'document':
            msg = self.telegram.sendDocument(chat_id=chat_id,
                                             document=content,
                                             **kwargs)
        elif msg_type == 'photo':
            msg = self.telegram.sendPhoto(chat_id=chat_id,
                                          photo=content,
                                          **kwargs)

        elif msg_type == 'audio':
            msg = self.telegram.sendAudio(chat_id=chat_id,
                                          audio=content,
                                          **kwargs)

        elif msg_type == 'video':
            msg = self.telegram.sendVideo(chat_id=chat_id,
                                          video=content,
                                          **kwargs)
        elif msg_type == 'sticker':
            msg = self.telegram.sendSticker(chat_id=chat_id,
                                            sticker=content,
                                            **kwargs)
        elif msg_type == 'location':
            msg = self.telegram.sendLocation(chat_id=chat_id,
                                             latitude=kwargs.pop('latitude', ''),
                                             longitude=kwargs.pop('longitude', ''),
                                             **kwargs)
        else:
            raise ValueError(f'Expected a valid choice for `msg_type`, got: {msg_type}.')
        return msg

    def _telegram_upload_stream(self, stream, **kwargs):
        """Perform upload defined in a stream."""
        msg = None
        try:
            stream.accept()
            msg = self._telegram_special_message(chat_id=stream.identifier.id,
                                                 content=stream.raw,
                                                 msg_type=stream.stream_type,
                                                 **kwargs)
        except Exception:
            log.exception(f'Upload of {stream.name} to {stream.identifier} failed.')
        else:
            if msg is None:
                stream.error()
            else:
                stream.success()

    def send_stream_request(self, identifier, fsource, name='file', size=None, stream_type=None):
        """Starts a file transfer.

        :param identifier: TelegramPerson or TelegramMUCOccupant
            Identifier of the Person or Room to send the stream to.

        :param fsource: str, dict or binary data
            File URL or binary content from a local file.
            Optionally a dict with binary content plus metadata can be given.
            See `stream_type` for more details.

        :param name: str, optional
            Name of the file. Not sure if this works always.

        :param size: str, optional
            Size of the file obtained with os.path.getsize.
            This is only used for debug logging purposes.

        :param stream_type: str, optional
            Type of the stream. Choices: 'document', 'photo', 'audio', 'video', 'sticker', 'location'.

            If 'video', a dict is optional as {'content': fsource, 'duration': str}.
            If 'voice', a dict is optional as {'content': fsource, 'duration': str}.
            If 'audio', a dict is optional as {'content': fsource, 'duration': str, 'performer': str, 'title': str}.

            For 'location' a dict is mandatory as {'latitude': str, 'longitude': str}.
            For 'venue': TODO # see: https://core.telegram.org/bots/api#sendvenue

        :return stream: str or Stream
            If `fsource` is str will return str, else return Stream.
        """

        def _telegram_metadata(fsource):
            if isinstance(fsource, dict):
                return fsource.pop('content'), fsource
            else:
                return fsource, None

        def _is_valid_url(url):
            try:
                from urlparse import urlparse
            except Exception:
                from urllib.parse import urlparse

            return bool(urlparse(url).scheme)

        content, meta = _telegram_metadata(fsource)
        if isinstance(content, str):
            if not _is_valid_url(content):
                raise ValueError(f'Not valid URL: {content}')

            self._telegram_special_message(chat_id=identifier.id,
                                           content=content,
                                           msg_type=stream_type,
                                           **meta)
            log.debug('Requesting upload of %s to %s (size hint: %d, stream type: %s).',
                      name, identifier.username, size, stream_type)

            stream = content
        else:
            stream = Stream(identifier, content, name, size, stream_type)
            log.debug('Requesting upload of %s to %s (size hint: %d, stream type: %s)',
                      name, identifier, size, stream_type)
            self.thread_pool.apply_async(self._telegram_upload_stream, (stream,))

        return stream

    @staticmethod
    def _is_numeric(input_):
        """Return true if input is a number"""
        try:
            int(input_)
            return True
        except ValueError:
            return False



# coding=utf-8
from os import path

import pytest

extra_plugin_dir = path.join(path.dirname(path.realpath(__file__)), 'test_link')


def test_linked_plugin_here(testbot):
    testbot.push_message('!status plugins')
    assert 'Dummy' in testbot.pop_message()



from __future__ import absolute_import
from errbot import BotPlugin, botcmd


class FlowTest(BotPlugin):
    """A plugin to test the flows
    see flowtest.png for the structure.
    """
    @botcmd
    def a(self, msg, args):
        return 'a'

    @botcmd
    def b(self, msg, args):
        return 'b'

    @botcmd
    def c(self, msg, args):
        return 'c'

    @botcmd(flow_only=True)
    def d(self, msg, args):
        return 'd'

    @botcmd
    def e(self, msg, args):
        return 'e'



from errbot import BotPlugin, cmdfilter


class TestCommandNotFoundFilter(BotPlugin):
    @cmdfilter(catch_unprocessed=True)
    def command_not_found(self, msg, cmd, args, dry_run, emptycmd=False):
        if not emptycmd:
            return msg, cmd, args

        return "Command fell through: {}".format(msg)



from os import path

extra_plugin_dir = path.join(path.dirname(path.realpath(__file__)), 'mention_plugin')


def test_foreign_mention(testbot):
    assert 'Somebody mentioned toto!' in testbot.exec_command('I am telling you something @toto')


def test_testbot_mention(testbot):
    assert 'Somebody mentioned me!' in testbot.exec_command('I am telling you something @Err')


def test_multiple_mentions(testbot):
    assert 'Somebody mentioned toto,titi!' in testbot.exec_command('I am telling you something @toto and @titi')



from os import path

extra_plugin_dir = path.join(path.dirname(path.realpath(__file__)), 'multi_plugin')

# This tests the decorellation between plugin class names and real names
# by making 2 instances of the same plugin collide on purpose.


def test_first(testbot):
    r = testbot.exec_command('!myname')
    assert 'Multi1' == r or 'Multi2' == r


def test_second(testbot):
    assert 'Multi2' == testbot.exec_command('!multi2-myname') or 'Multi1' == testbot.exec_command('!multi1-myname')



from __future__ import absolute_import
from errbot import botflow, FlowRoot, BotFlow


class FlowDefinitions(BotFlow):
    """A plugin to test the flows
    see flowtest.png for the structure.
    """
    @botflow
    def w1(self, flow: FlowRoot):
        "documentation of W1"
        a_node = flow.connect('a')   # no autotrigger
        b_node = a_node.connect('b')
        c_node = a_node.connect('c')  # crosses the autotrigger of w2
        d_node = c_node.connect('d')

        assert a_node.hints
        assert b_node.hints
        assert c_node.hints
        assert d_node.hints

    @botflow
    def w2(self, flow: FlowRoot):
        """documentation of W2"""
        c_node = flow.connect('c', auto_trigger=True)
        b_node = c_node.connect('b')
        e_node = flow.connect('e', auto_trigger=True)  # 2 autotriggers for the same workflow
        d_node = e_node.connect('d')

    @botflow
    def w3(self, flow: FlowRoot):
        """documentation of W3"""
        c_node = flow.connect('a', room_flow=True)
        b_node = c_node.connect('b')

    @botflow
    def w4(self, flow: FlowRoot):
        """documentation of W4"""
        a_node = flow.connect('a')
        b_node = a_node.connect('b')
        c_node = b_node.connect('c')
        c_node.connect('d')

        # set some hinting.
        flow.hints = False
        a_node.hints = False
        b_node.hints = True
        c_node.hint = False



from os import path

extra_plugin_dir = path.join(path.dirname(path.realpath(__file__)), 'dyna_plugin')


def test_simple(testbot):
    assert 'added' in testbot.exec_command('!add_simple')
    assert 'yep' in testbot.exec_command('!say_yep')
    assert 'foo' in testbot.exec_command('!say_foo')
    assert 'documented' in testbot.exec_command('!help')
    assert 'removed' in testbot.exec_command('!remove_simple')
    assert 'Command "say_foo" not found' in testbot.exec_command('!say_foo')


def test_arg(testbot):
    assert 'added' in testbot.exec_command('!add_arg')
    assert 'string to echo is string_to_echo' in testbot.exec_command('!echo_to_me string_to_echo')
    assert 'removed' in testbot.exec_command('!remove_arg')
    assert ('Command "echo_to_me" / "echo_to_me string_to_echo" not found'
            in testbot.exec_command('!echo_to_me string_to_echo'))


def test_re(testbot):
    assert 'added' in testbot.exec_command('!add_re')
    assert 'fffound' in testbot.exec_command('I said cheese')
    assert 'removed' in testbot.exec_command('!remove_re')


def test_saw(testbot):
    assert 'added' in testbot.exec_command('!add_saw')
    assert 'foo+bar+baz' in testbot.exec_command('!splitme foo,bar,baz')
    assert 'removed' in testbot.exec_command('!remove_saw')


def test_clashing(testbot):
    assert 'original' in testbot.exec_command('!clash')
    assert 'clashing.clash clashes with Dyna.clash so it has been renamed clashing-clash' in \
           testbot.exec_command('!add_clashing')
    assert 'added' in testbot.pop_message()
    assert 'original' in testbot.exec_command('!clash')
    assert 'dynamic' in testbot.exec_command('!clashing-clash')
    assert 'removed' in testbot.exec_command('!remove_clashing')
    assert 'original' in testbot.exec_command('!clash')
    assert 'not found' in testbot.exec_command('!clashing-clash')



import logging
import os
from errbot.plugin_info import PluginInfo
from jinja2 import Environment, FileSystemLoader
from pathlib import Path

log = logging.getLogger(__name__)


def make_templates_path(root: Path) -> Path:
    return root / 'templates'


system_templates_path = str(make_templates_path(Path(__file__).parent))
template_path = [system_templates_path]
env = Environment(loader=FileSystemLoader(template_path),
                  trim_blocks=True,
                  keep_trailing_newline=False,
                  autoescape=True)


def tenv():
    return env


def add_plugin_templates_path(plugin_info: PluginInfo):
    global env
    tmpl_path = make_templates_path(plugin_info.location.parent)
    if tmpl_path.exists():
        log.debug('Templates directory found for this plugin [%s]', tmpl_path)
        template_path.append(str(tmpl_path))  # for webhooks

        # Ditch and recreate a new templating environment
        env = Environment(loader=FileSystemLoader(template_path), autoescape=True)
        return
    log.debug('No templates directory found for this plugin [Looking for %s]', tmpl_path)


def remove_plugin_templates_path(plugin_info: PluginInfo):
    global env
    tmpl_path = str(make_templates_path(plugin_info.location.parent))
    if tmpl_path in template_path:
        template_path.remove(tmpl_path)
        # Ditch and recreate a new templating environment
        env = Environment(loader=FileSystemLoader(template_path), autoescape=True)



# -*- coding=utf-8 -*-
from errbot import BotPlugin, botcmd


class I18nTest(BotPlugin):
    """A Just a test plugin to see if it is picked up.
    """
    @botcmd
    def i18n_1(self, msg, args):
        return 'язы́к'

    @botcmd(name='ру́сский')
    def i18n_2(self, msg, args):
        return 'OK'

    @botcmd(name='prefix_ру́сский')
    def i18n_3(self, msg, args):
        return 'OK'

    @botcmd(name='ру́сский_suffix')
    def i18n_4(self, msg, args):
        return 'OK'



import collections
import copyreg
import json
import logging
import re
import sys
import pprint
from functools import lru_cache
from typing import BinaryIO

from markdown import Markdown
from markdown.extensions.extra import ExtraExtension
from markdown.preprocessors import Preprocessor

from errbot.backends.base import Identifier, Message, Presence, ONLINE, AWAY, Room, RoomError, RoomDoesNotExistError, \
    UserDoesNotExistError, RoomOccupant, Person, Card, Stream
from errbot.core import ErrBot
from errbot.utils import split_string_after
from errbot.rendering.ansiext import AnsiExtension, enable_format, IMTEXT_CHRS

log = logging.getLogger(__name__)


try:
    from slackclient import SlackClient
except ImportError:
    log.exception("Could not start the Slack back-end")
    log.fatal(
        "You need to install the slackclient support in order to use the Slack backend.\n"
        "You can do `pip install errbot[slack]` to install it"
    )
    sys.exit(1)

# The Slack client automatically turns a channel name into a clickable
# link if you prefix it with a #. Other clients receive this link as a
# token matching this regex.
SLACK_CLIENT_CHANNEL_HYPERLINK = re.compile(r'^<#(?P<id>(C|G)[0-9A-Z]+)>$')

# Empirically determined message size limit.
SLACK_MESSAGE_LIMIT = 4096

USER_IS_BOT_HELPTEXT = (
    "Connected to Slack using a bot account, which cannot manage "
    "channels itself (you must invite the bot to channels instead, "
    "it will auto-accept) nor invite people.\n\n"
    "If you need this functionality, you will have to create a "
    "regular user account and connect Errbot using that account. "
    "For this, you will also need to generate a user token at "
    "https://api.slack.com/web."
)

COLORS = {
    'red': '#FF0000',
    'green': '#008000',
    'yellow': '#FFA500',
    'blue': '#0000FF',
    'white': '#FFFFFF',
    'cyan': '#00FFFF'
}  # Slack doesn't know its colors


MARKDOWN_LINK_REGEX = re.compile(r'(?<!!)\[(?P<text>[^\]]+?)\]\((?P<uri>[a-zA-Z0-9]+?:\S+?)\)')


def slack_markdown_converter(compact_output=False):
    """
    This is a Markdown converter for use with Slack.
    """
    enable_format('imtext', IMTEXT_CHRS, borders=not compact_output)
    md = Markdown(output_format='imtext', extensions=[ExtraExtension(), AnsiExtension()])
    md.preprocessors['LinkPreProcessor'] = LinkPreProcessor(md)
    md.stripTopLevelTags = False
    return md


class LinkPreProcessor(Preprocessor):
    """
    This preprocessor converts markdown URL notation into Slack URL notation
    as described at https://api.slack.com/docs/formatting, section "Linking to URLs".
    """
    def run(self, lines):
        for i, line in enumerate(lines):
            lines[i] = MARKDOWN_LINK_REGEX.sub(r'&lt;\2|\1&gt;', line)
        return lines


class SlackAPIResponseError(RuntimeError):
    """Slack API returned a non-OK response"""

    def __init__(self, *args, error='', **kwargs):
        """
        :param error:
            The 'error' key from the API response data
        """
        self.error = error
        super().__init__(*args, **kwargs)


class SlackPerson(Person):
    """
    This class describes a person on Slack's network.
    """

    def __init__(self, sc, userid=None, channelid=None):
        if userid is not None and userid[0] not in ('U', 'B', 'W'):
            raise Exception(f'This is not a Slack user or bot id: {userid} (should start with U, B or W)')

        if channelid is not None and channelid[0] not in ('D', 'C', 'G'):
            raise Exception(f'This is not a valid Slack channelid: {channelid} (should start with D, C or G)')

        self._userid = userid
        self._channelid = channelid
        self._sc = sc

    @property
    def userid(self):
        return self._userid

    @property
    def username(self):
        """Convert a Slack user ID to their user name"""
        user = self._sc.server.users.find(self._userid)
        if user is None:
            log.error('Cannot find user with ID %s', self._userid)
            return f'<{self._userid}>'
        return user.name

    @property
    def channelid(self):
        return self._channelid

    @property
    def channelname(self):
        """Convert a Slack channel ID to its channel name"""
        if self._channelid is None:
            return None

        channel = self._sc.server.channels.find(self._channelid)
        if channel is None:
            raise RoomDoesNotExistError(f'No channel with ID {self._channelid} exists.')
        return channel.name

    @property
    def domain(self):
        return self._sc.server.domain

    # Compatibility with the generic API.
    client = channelid
    nick = username

    # Override for ACLs
    @property
    def aclattr(self):
        # Note: Don't use str(self) here because that will return
        # an incorrect format from SlackMUCOccupant.
        return f'@{self.username}'

    @property
    def fullname(self):
        """Convert a Slack user ID to their user name"""
        user = self._sc.server.users.find(self._userid)
        if user is None:
            log.error('Cannot find user with ID %s', self._userid)
            return f'<{self._userid}>'
        return user.real_name

    def __unicode__(self):
        return f'@{self.username}'

    def __str__(self):
        return self.__unicode__()

    def __eq__(self, other):
        if not isinstance(other, SlackPerson):
            log.warning('tried to compare a SlackPerson with a %s', type(other))
            return False
        return other.userid == self.userid

    def __hash__(self):
        return self.userid.__hash__()

    @property
    def person(self):
        # Don't use str(self) here because we want SlackRoomOccupant
        # to return just our @username too.
        return f'@{self.username}'


class SlackRoomOccupant(RoomOccupant, SlackPerson):
    """
    This class represents a person inside a MUC.
    """
    def __init__(self, sc, userid, channelid, bot):
        super().__init__(sc, userid, channelid)
        self._room = SlackRoom(channelid=channelid, bot=bot)

    @property
    def room(self):
        return self._room

    def __unicode__(self):
        return f'#{self._room.name}/{self.username}'

    def __str__(self):
        return self.__unicode__()

    def __eq__(self, other):
        if not isinstance(other, RoomOccupant):
            log.warning('tried to compare a SlackRoomOccupant with a SlackPerson %s vs %s', self, other)
            return False
        return other.room.id == self.room.id and other.userid == self.userid


class SlackBot(SlackPerson):
    """
    This class describes a bot on Slack's network.
    """
    def __init__(self, sc, bot_id, bot_username):
        self._bot_id = bot_id
        self._bot_username = bot_username
        super().__init__(sc=sc, userid=bot_id)

    @property
    def username(self):
        return self._bot_username

    # Beware of gotcha. Without this, nick would point to username of SlackPerson.
    nick = username

    @property
    def aclattr(self):
        # Make ACLs match against integration ID rather than human-readable
        # nicknames to avoid webhooks impersonating other people.
        return f'<{self._bot_id}>'

    @property
    def fullname(self):
        return None


class SlackRoomBot(RoomOccupant, SlackBot):
    """
    This class represents a bot inside a MUC.
    """
    def __init__(self, sc, bot_id, bot_username, channelid, bot):
        super().__init__(sc, bot_id, bot_username)
        self._room = SlackRoom(channelid=channelid, bot=bot)

    @property
    def room(self):
        return self._room

    def __unicode__(self):
        return f'#{self._room.name}/{self.username}'

    def __str__(self):
        return self.__unicode__()

    def __eq__(self, other):
        if not isinstance(other, RoomOccupant):
            log.warning('tried to compare a SlackRoomBotOccupant with a SlackPerson %s vs %s', self, other)
            return False
        return other.room.id == self.room.id and other.userid == self.userid


class SlackBackend(ErrBot):

    @staticmethod
    def _unpickle_identifier(identifier_str):
        return SlackBackend.__build_identifier(identifier_str)

    @staticmethod
    def _pickle_identifier(identifier):
        return SlackBackend._unpickle_identifier, (str(identifier),)

    def _register_identifiers_pickling(self):
        """
        Register identifiers pickling.

        As Slack needs live objects in its identifiers, we need to override their pickling behavior.
        But for the unpickling to work we need to use bot.build_identifier, hence the bot parameter here.
        But then we also need bot for the unpickling so we save it here at module level.
        """
        SlackBackend.__build_identifier = self.build_identifier
        for cls in (SlackPerson, SlackRoomOccupant, SlackRoom):
            copyreg.pickle(cls, SlackBackend._pickle_identifier, SlackBackend._unpickle_identifier)

    def __init__(self, config):
        super().__init__(config)
        identity = config.BOT_IDENTITY
        self.token = identity.get('token', None)
        self.proxies = identity.get('proxies', None)
        if not self.token:
            log.fatal(
                'You need to set your token (found under "Bot Integration" on Slack) in '
                'the BOT_IDENTITY setting in your configuration. Without this token I '
                'cannot connect to Slack.'
            )
            sys.exit(1)
        self.sc = None  # Will be initialized in serve_once
        compact = config.COMPACT_OUTPUT if hasattr(config, 'COMPACT_OUTPUT') else False
        self.md = slack_markdown_converter(compact)
        self._register_identifiers_pickling()

    def api_call(self, method, data=None, raise_errors=True):
        """
        Make an API call to the Slack API and return response data.

        This is a thin wrapper around `SlackClient.server.api_call`.

        :param method:
            The API method to invoke (see https://api.slack.com/methods/).
        :param raise_errors:
            Whether to raise :class:`~SlackAPIResponseError` if the API
            returns an error
        :param data:
            A dictionary with data to pass along in the API request.
        :returns:
            A dictionary containing the (JSON-decoded) API response
        :raises:
            :class:`~SlackAPIResponseError` if raise_errors is True and the
            API responds with `{"ok": false}`
        """
        if data is None:
            data = {}
        response = self.sc.api_call(method, **data)
        if not isinstance(response, collections.Mapping):
            # Compatibility with SlackClient < 1.0.0
            response = json.loads(response.decode('utf-8'))

        if raise_errors and not response['ok']:
            raise SlackAPIResponseError(f"Slack API call to {method} failed: {response['error']}",
                                        error=response['error'])
        return response

    def update_alternate_prefixes(self):
        """Converts BOT_ALT_PREFIXES to use the slack ID instead of name

        Slack only acknowledges direct callouts `@username` in chat if referred
        by using the ID of that user.
        """
        # convert BOT_ALT_PREFIXES to a list
        try:
            bot_prefixes = self.bot_config.BOT_ALT_PREFIXES.split(',')
        except AttributeError:
            bot_prefixes = list(self.bot_config.BOT_ALT_PREFIXES)

        converted_prefixes = []
        for prefix in bot_prefixes:
            try:
                converted_prefixes.append(f'<@{self.username_to_userid(prefix)}>')
            except Exception as e:
                log.error('Failed to look up Slack userid for alternate prefix "%s": %s', prefix, e)

        self.bot_alt_prefixes = tuple(x.lower() for x in self.bot_config.BOT_ALT_PREFIXES)
        log.debug('Converted bot_alt_prefixes: %s', self.bot_config.BOT_ALT_PREFIXES)

    def serve_once(self):
        self.sc = SlackClient(self.token, proxies=self.proxies)

        log.info('Verifying authentication token')
        self.auth = self.api_call("auth.test", raise_errors=False)
        if not self.auth['ok']:
            raise SlackAPIResponseError(error=f"Couldn't authenticate with Slack. Server said: {self.auth['error']}")
        log.debug("Token accepted")
        self.bot_identifier = SlackPerson(self.sc, self.auth["user_id"])

        log.info("Connecting to Slack real-time-messaging API")
        if self.sc.rtm_connect():
            log.info("Connected")
            # Block on reads instead of using the busy loop suggested in slackclient docs
            # https://github.com/slackapi/python-slackclient/issues/46#issuecomment-165674808
            self.sc.server.websocket.sock.setblocking(True)
            self.reset_reconnection_count()

            # Inject bot identity to alternative prefixes
            self.update_alternate_prefixes()

            try:
                while True:
                    for message in self.sc.rtm_read():
                        self._dispatch_slack_message(message)
            except KeyboardInterrupt:
                log.info("Interrupt received, shutting down..")
                return True
            except Exception:
                log.exception("Error reading from RTM stream:")
            finally:
                log.debug("Triggering disconnect callback")
                self.disconnect_callback()
        else:
            raise Exception('Connection failed, invalid token ?')

    def _dispatch_slack_message(self, message):
        """
        Process an incoming message from slack.

        """
        if 'type' not in message:
            log.debug("Ignoring non-event message: %s.", message)
            return

        event_type = message['type']

        event_handlers = {
            'hello': self._hello_event_handler,
            'presence_change': self._presence_change_event_handler,
            'message': self._message_event_handler,
            'member_joined_channel': self._member_joined_channel_event_handler,
        }

        event_handler = event_handlers.get(event_type)

        if event_handler is None:
            log.debug('No event handler available for %s, ignoring this event', event_type)
            return
        try:
            log.debug('Processing slack event: %s', message)
            event_handler(message)
        except Exception:
            log.exception(f'{event_type} event handler raised an exception')

    def _hello_event_handler(self, event):
        """Event handler for the 'hello' event"""
        self.connect_callback()
        self.callback_presence(Presence(identifier=self.bot_identifier, status=ONLINE))

    def _presence_change_event_handler(self, event):
        """Event handler for the 'presence_change' event"""

        idd = SlackPerson(self.sc, event['user'])
        presence = event['presence']
        # According to https://api.slack.com/docs/presence, presence can
        # only be one of 'active' and 'away'
        if presence == 'active':
            status = ONLINE
        elif presence == 'away':
            status = AWAY
        else:
            log.error(f'It appears the Slack API changed, I received an unknown presence type {presence}.')
            status = ONLINE
        self.callback_presence(Presence(identifier=idd, status=status))

    def _message_event_handler(self, event):
        """Event handler for the 'message' event"""
        channel = event['channel']
        if channel[0] not in 'CGD':
            log.warning("Unknown message type! Unable to handle %s", channel)
            return

        subtype = event.get('subtype', None)

        if subtype in ("message_deleted", "channel_topic", "message_replied"):
            log.debug("Message of type %s, ignoring this event", subtype)
            return

        if subtype == "message_changed" and 'attachments' in event['message']:
            # If you paste a link into Slack, it does a call-out to grab details
            # from it so it can display this in the chatroom. These show up as
            # message_changed events with an 'attachments' key in the embedded
            # message. We should completely ignore these events otherwise we
            # could end up processing bot commands twice (user issues a command
            # containing a link, it gets processed, then Slack triggers the
            # message_changed event and we end up processing it again as a new
            # message. This is not what we want).
            log.debug(
                "Ignoring message_changed event with attachments, likely caused "
                "by Slack auto-expanding a link"
            )
            return

        if 'message' in event:
            text = event['message'].get('text', '')
            user = event['message'].get('user', event.get('bot_id'))
        else:
            text = event.get('text', '')
            user = event.get('user', event.get('bot_id'))

        text, mentioned = self.process_mentions(text)

        text = self.sanitize_uris(text)

        log.debug('Saw an event: %s', pprint.pformat(event))
        log.debug('Escaped IDs event text: %s', text)

        msg = Message(
            text,
            extras={
                'attachments': event.get('attachments'),
                'slack_event': event,
            },
        )

        if channel.startswith('D'):
            if subtype == "bot_message":
                msg.frm = SlackBot(
                    self.sc,
                    bot_id=event.get('bot_id'),
                    bot_username=event.get('username', '')
                )
            else:
                msg.frm = SlackPerson(self.sc, user, event['channel'])
            msg.to = SlackPerson(self.sc, self.username_to_userid(self.sc.server.username),
                                 event['channel'])
            channel_link_name = event['channel']
        else:
            if subtype == "bot_message":
                msg.frm = SlackRoomBot(
                    self.sc,
                    bot_id=event.get('bot_id'),
                    bot_username=event.get('username', ''),
                    channelid=event['channel'],
                    bot=self
                )
            else:
                msg.frm = SlackRoomOccupant(self.sc, user, event['channel'], bot=self)
            msg.to = SlackRoom(channelid=event['channel'], bot=self)
            channel_link_name = msg.to.name

        msg.extras['url'] = f'https://{self.sc.server.domain}.slack.com/archives/' \
                            f'{channel_link_name}/p{self._ts_for_message(msg).replace(".", "")}'

        self.callback_message(msg)

        if mentioned:
            self.callback_mention(msg, mentioned)

    def _member_joined_channel_event_handler(self, event):
        """Event handler for the 'member_joined_channel' event"""
        user = SlackPerson(self.sc, event['user'])
        if user == self.bot_identifier:
            self.callback_room_joined(SlackRoom(channelid=event['channel'], bot=self))

    def userid_to_username(self, id_):
        """Convert a Slack user ID to their user name"""
        user = self.sc.server.users.get(id_)
        if user is None:
            raise UserDoesNotExistError(f'Cannot find user with ID {id_}.')
        return user.name

    def username_to_userid(self, name):
        """Convert a Slack user name to their user ID"""
        name = name.lstrip('@')
        user = self.sc.server.users.find(name)
        if user is None:
            raise UserDoesNotExistError(f'Cannot find user {name}.')
        return user.id

    def channelid_to_channelname(self, id_):
        """Convert a Slack channel ID to its channel name"""
        channel = [channel for channel in self.sc.server.channels if channel.id == id_]
        if not channel:
            raise RoomDoesNotExistError(f'No channel with ID {id_} exists.')
        return channel[0].name

    def channelname_to_channelid(self, name):
        """Convert a Slack channel name to its channel ID"""
        name = name.lstrip('#')
        channel = [channel for channel in self.sc.server.channels if channel.name == name]
        if not channel:
            raise RoomDoesNotExistError(f'No channel named {name} exists')
        return channel[0].id

    def channels(self, exclude_archived=True, joined_only=False):
        """
        Get all channels and groups and return information about them.

        :param exclude_archived:
            Exclude archived channels/groups
        :param joined_only:
            Filter out channels the bot hasn't joined
        :returns:
            A list of channel (https://api.slack.com/types/channel)
            and group (https://api.slack.com/types/group) types.

        See also:
          * https://api.slack.com/methods/channels.list
          * https://api.slack.com/methods/groups.list
        """
        response = self.api_call('channels.list', data={'exclude_archived': exclude_archived})
        channels = [channel for channel in response['channels']
                    if channel['is_member'] or not joined_only]

        response = self.api_call('groups.list', data={'exclude_archived': exclude_archived})
        # No need to filter for 'is_member' in this next call (it doesn't
        # (even exist) because leaving a group means you have to get invited
        # back again by somebody else.
        groups = [group for group in response['groups']]

        return channels + groups

    @lru_cache(1024)
    def get_im_channel(self, id_):
        """Open a direct message channel to a user"""
        try:
            response = self.api_call('im.open', data={'user': id_})
            return response['channel']['id']
        except SlackAPIResponseError as e:
            if e.error == "cannot_dm_bot":
                log.info('Tried to DM a bot.')
                return None
            else:
                raise e

    def _prepare_message(self, msg):  # or card
        """
        Translates the common part of messaging for Slack.
        :param msg: the message you want to extract the Slack concept from.
        :return: a tuple to user human readable, the channel id
        """
        if msg.is_group:
            to_channel_id = msg.to.id
            to_humanreadable = msg.to.name if msg.to.name else self.channelid_to_channelname(to_channel_id)
        else:
            to_humanreadable = msg.to.username
            to_channel_id = msg.to.channelid
            if to_channel_id.startswith('C'):
                log.debug("This is a divert to private message, sending it directly to the user.")
                to_channel_id = self.get_im_channel(self.username_to_userid(msg.to.username))
        return to_humanreadable, to_channel_id

    def send_message(self, msg):
        super().send_message(msg)

        if msg.parent is not None:
            # we are asked to reply to a specify thread.
            try:
                msg.extras['thread_ts'] = self._ts_for_message(msg.parent)
            except KeyError:
                # Gives to the user a more interesting explanation if we cannot find a ts from the parent.
                log.exception('The provided parent message is not a Slack message '
                              'or does not contain a Slack timestamp.')

        to_humanreadable = "<unknown>"
        try:
            if msg.is_group:
                to_channel_id = msg.to.id
                to_humanreadable = msg.to.name if msg.to.name else self.channelid_to_channelname(to_channel_id)
            else:
                to_humanreadable = msg.to.username
                if isinstance(msg.to, RoomOccupant):  # private to a room occupant -> this is a divert to private !
                    log.debug("This is a divert to private message, sending it directly to the user.")
                    to_channel_id = self.get_im_channel(self.username_to_userid(msg.to.username))
                else:
                    to_channel_id = msg.to.channelid

            msgtype = "direct" if msg.is_direct else "channel"
            log.debug('Sending %s message to %s (%s).', msgtype, to_humanreadable, to_channel_id)
            body = self.md.convert(msg.body)
            log.debug('Message size: %d.', len(body))

            limit = min(self.bot_config.MESSAGE_SIZE_LIMIT, SLACK_MESSAGE_LIMIT)
            parts = self.prepare_message_body(body, limit)

            timestamps = []
            for part in parts:
                data = {
                    'channel': to_channel_id,
                    'text': part,
                    'unfurl_media': 'true',
                    'link_names': '1',
                    'as_user': 'true',
                }

                # Keep the thread_ts to answer to the same thread.
                if 'thread_ts' in msg.extras:
                    data['thread_ts'] = msg.extras['thread_ts']

                result = self.api_call('chat.postMessage', data=data)
                timestamps.append(result['ts'])

            msg.extras['ts'] = timestamps
        except Exception:
            log.exception(f'An exception occurred while trying to send the following message '
                          f'to {to_humanreadable}: {msg.body}.')

    def _slack_upload(self, stream: Stream) -> None:
        """
        Performs an upload defined in a stream
        :param stream: Stream object
        :return: None
        """
        try:
            stream.accept()
            resp = self.api_call('files.upload', data={
                'channels': stream.identifier.channelid,
                'filename': stream.name,
                'file': stream
            })
            if 'ok' in resp and resp['ok']:
                stream.success()
            else:
                stream.error()
        except Exception:
            log.exception(f'Upload of {stream.name} to {stream.identifier.channelname} failed.')

    def send_stream_request(self,
                            user: Identifier,
                            fsource: BinaryIO,
                            name: str = None,
                            size: int = None,
                            stream_type: str = None) -> Stream:
        """
            Starts a file transfer. For Slack, the size and stream_type are unsupported

            :param user: is the identifier of the person you want to send it to.
            :param fsource: is a file object you want to send.
            :param name: is an optional filename for it.
            :param size: not supported in Slack backend
            :param stream_type: not supported in Slack backend

            :return Stream: object on which you can monitor the progress of it.
        """
        stream = Stream(user, fsource, name, size, stream_type)
        log.debug('Requesting upload of %s to %s (size hint: %d, stream type: %s).',
                  name, user.channelname, size, stream_type)
        self.thread_pool.apply_async(self._slack_upload, (stream,))
        return stream

    def send_card(self, card: Card):
        if isinstance(card.to, RoomOccupant):
            card.to = card.to.room
        to_humanreadable, to_channel_id = self._prepare_message(card)
        attachment = {}
        if card.summary:
            attachment['pretext'] = card.summary
        if card.title:
            attachment['title'] = card.title
        if card.link:
            attachment['title_link'] = card.link
        if card.image:
            attachment['image_url'] = card.image
        if card.thumbnail:
            attachment['thumb_url'] = card.thumbnail

        if card.color:
            attachment['color'] = COLORS[card.color] if card.color in COLORS else card.color

        if card.fields:
            attachment['fields'] = [{'title': key, 'value': value, 'short': True} for key, value in card.fields]

        limit = min(self.bot_config.MESSAGE_SIZE_LIMIT, SLACK_MESSAGE_LIMIT)
        parts = self.prepare_message_body(card.body, limit)
        part_count = len(parts)
        footer = attachment.get('footer', '')
        for i in range(part_count):
            if part_count > 1:
                attachment['footer'] = f'{footer} [{i + 1}/{part_count}]'
            attachment['text'] = parts[i]
            data = {
                'channel': to_channel_id,
                'attachments': json.dumps([attachment]),
                'link_names': '1',
                'as_user': 'true'
            }
            try:
                log.debug('Sending data:\n%s', data)
                self.api_call('chat.postMessage', data=data)
            except Exception:
                log.exception(f'An exception occurred while trying to send a card to {to_humanreadable}.[{card}]')

    def __hash__(self):
        return 0  # this is a singleton anyway

    def change_presence(self, status: str = ONLINE, message: str = '') -> None:
        self.api_call('users.setPresence', data={'presence': 'auto' if status == ONLINE else 'away'})

    @staticmethod
    def prepare_message_body(body, size_limit):
        """
        Returns the parts of a message chunked and ready for sending.

        This is a staticmethod for easier testing.

        Args:
            body (str)
            size_limit (int): chunk the body into sizes capped at this maximum

        Returns:
            [str]

        """
        fixed_format = body.startswith('```')  # hack to fix the formatting
        parts = list(split_string_after(body, size_limit))

        if len(parts) == 1:
            # If we've got an open fixed block, close it out
            if parts[0].count('```') % 2 != 0:
                parts[0] += '\n```\n'
        else:
            for i, part in enumerate(parts):
                starts_with_code = part.startswith('```')

                # If we're continuing a fixed block from the last part
                if fixed_format and not starts_with_code:
                    parts[i] = '```\n' + part

                # If we've got an open fixed block, close it out
                if part.count('```') % 2 != 0:
                    parts[i] += '\n```\n'

        return parts

    @staticmethod
    def extract_identifiers_from_string(text):
        """
        Parse a string for Slack user/channel IDs.

        Supports strings with the following formats::

            <#C12345>
            <@U12345>
            <@U12345|user>
            @user
            #channel/user
            #channel

        Returns the tuple (username, userid, channelname, channelid).
        Some elements may come back as None.
        """
        exception_message = (
            'Unparseable slack identifier, should be of the format `<#C12345>`, `<@U12345>`, '
            '`<@U12345|user>`, `@user`, `#channel/user` or `#channel`. (Got `%s`)'
        )
        text = text.strip()

        if text == '':
            raise ValueError(exception_message % '')

        channelname = None
        username = None
        channelid = None
        userid = None

        if text[0] == '<' and text[-1] == '>':
            exception_message = 'Unparseable slack ID, should start with U, B, C, G, D or W (got `%s`)'
            text = text[2:-1]
            if text == '':
                raise ValueError(exception_message % '')
            if text[0] in ('U', 'B', 'W'):
                if '|' in text:
                    userid, username = text.split('|')
                else:
                    userid = text
            elif text[0] in ('C', 'G', 'D'):
                channelid = text
            else:
                raise ValueError(exception_message % text)
        elif text[0] == '@':
            username = text[1:]
        elif text[0] == '#':
            plainrep = text[1:]
            if '/' in text:
                channelname, username = plainrep.split('/', 1)
            else:
                channelname = plainrep
        else:
            raise ValueError(exception_message % text)

        return username, userid, channelname, channelid

    def build_identifier(self, txtrep):
        """
        Build a :class:`SlackIdentifier` from the given string txtrep.

        Supports strings with the formats accepted by
        :func:`~extract_identifiers_from_string`.
        """
        log.debug('building an identifier from %s.', txtrep)
        username, userid, channelname, channelid = self.extract_identifiers_from_string(txtrep)

        if userid is None and username is not None:
            userid = self.username_to_userid(username)
        if channelid is None and channelname is not None:
            channelid = self.channelname_to_channelid(channelname)
        if userid is not None and channelid is not None:
            return SlackRoomOccupant(self.sc, userid, channelid, bot=self)
        if userid is not None:
            return SlackPerson(self.sc, userid, self.get_im_channel(userid))
        if channelid is not None:
            return SlackRoom(channelid=channelid, bot=self)

        raise Exception(
            "You found a bug. I expected at least one of userid, channelid, username or channelname "
            "to be resolved but none of them were. This shouldn't happen so, please file a bug."
        )

    def is_from_self(self, msg: Message) -> bool:
        return self.bot_identifier.userid == msg.frm.userid

    def build_reply(self, msg, text=None, private=False, threaded=False):
        response = self.build_message(text)

        if 'thread_ts' in msg.extras['slack_event']:
            # If we reply to a threaded message, keep it in the thread.
            response.extras['thread_ts'] = msg.extras['slack_event']['thread_ts']
        elif threaded:
            # otherwise check if we should start a new thread
            response.parent = msg

        response.frm = self.bot_identifier
        if private:
            response.to = msg.frm
        else:
            response.to = msg.frm.room if isinstance(msg.frm, RoomOccupant) else msg.frm
        return response

    def add_reaction(self, msg: Message, reaction: str) -> None:
        """
        Add the specified reaction to the Message if you haven't already.
        :param msg: A Message.
        :param reaction: A str giving an emoji, without colons before and after.
        :raises: ValueError if the emoji doesn't exist.
        """
        return self._react('reactions.add', msg, reaction)

    def remove_reaction(self, msg: Message, reaction: str) -> None:
        """
        Remove the specified reaction from the Message if it is currently there.
        :param msg: A Message.
        :param reaction: A str giving an emoji, without colons before and after.
        :raises: ValueError if the emoji doesn't exist.
        """
        return self._react('reactions.remove', msg, reaction)

    def _react(self, method: str, msg: Message, reaction: str) -> None:
        try:
            # this logic is from send_message
            if msg.is_group:
                to_channel_id = msg.to.id
            else:
                to_channel_id = msg.to.channelid

            ts = self._ts_for_message(msg)

            self.api_call(method, data={'channel': to_channel_id,
                                        'timestamp': ts,
                                        'name': reaction})
        except SlackAPIResponseError as e:
            if e.error == 'invalid_name':
                raise ValueError(e.error, 'No such emoji', reaction)
            elif e.error in ('no_reaction', 'already_reacted'):
                # This is common if a message was edited after you reacted to it, and you reacted to it again.
                # Chances are you don't care about this. If you do, call api_call() directly.
                pass
            else:
                raise SlackAPIResponseError(error=e.error)

    def _ts_for_message(self, msg):
        try:
            return msg.extras['slack_event']['message']['ts']
        except KeyError:
            return msg.extras['slack_event']['ts']

    def shutdown(self):
        super().shutdown()

    @property
    def mode(self):
        return 'slack'

    def query_room(self, room):
        """ Room can either be a name or a channelid """
        if room.startswith('C') or room.startswith('G'):
            return SlackRoom(channelid=room, bot=self)

        m = SLACK_CLIENT_CHANNEL_HYPERLINK.match(room)
        if m is not None:
            return SlackRoom(channelid=m.groupdict()['id'], bot=self)

        return SlackRoom(name=room, bot=self)

    def rooms(self):
        """
        Return a list of rooms the bot is currently in.

        :returns:
            A list of :class:`~SlackRoom` instances.
        """
        channels = self.channels(joined_only=True, exclude_archived=True)
        return [SlackRoom(channelid=channel['id'], bot=self) for channel in channels]

    def prefix_groupchat_reply(self, message, identifier):
        super().prefix_groupchat_reply(message, identifier)
        message.body = f'@{identifier.nick}: {message.body}'

    @staticmethod
    def sanitize_uris(text):
        """
        Sanitizes URI's present within a slack message. e.g.
        <mailto:example@example.org|example@example.org>,
        <http://example.org|example.org>
        <http://example.org>

        :returns:
            string
        """
        text = re.sub(r'<([^|>]+)\|([^|>]+)>', r'\2', text)
        text = re.sub(r'<(http([^>]+))>', r'\1', text)

        return text

    def process_mentions(self, text):
        """
        Process mentions in a given string
        :returns:
            A formatted string of the original message
            and a list of :class:`~SlackPerson` instances.
        """
        mentioned = []

        m = re.findall('<@[^>]*>*', text)

        for word in m:
            try:
                identifier = self.build_identifier(word)
            except Exception as e:
                log.debug("Tried to build an identifier from '%s' but got exception: %s", word, e)
                continue

            # We only track mentions of persons.
            if isinstance(identifier, SlackPerson):
                log.debug('Someone mentioned')
                mentioned.append(identifier)
                text = text.replace(word, str(identifier))

        return text, mentioned


class SlackRoom(Room):
    def __init__(self, name=None, channelid=None, bot=None):
        if channelid is not None and name is not None:
            raise ValueError("channelid and name are mutually exclusive")

        if name is not None:
            if name.startswith('#'):
                self._name = name[1:]
            else:
                self._name = name
        else:
            self._name = bot.channelid_to_channelname(channelid)

        self._id = None
        self._bot = bot
        self.sc = bot.sc

    def __str__(self):
        return f'#{self.name}'

    @property
    def channelname(self):
        return self._name

    @property
    def _channel(self):
        """
        The channel object exposed by SlackClient
        """
        id_ = self.sc.server.channels.find(self.name)
        if id_ is None:
            raise RoomDoesNotExistError(f"{str(self)} does not exist (or is a private group you don't have access to)")
        return id_

    @property
    def _channel_info(self):
        """
        Channel info as returned by the Slack API.

        See also:
          * https://api.slack.com/methods/channels.list
          * https://api.slack.com/methods/groups.list
        """
        if self.private:
            return self._bot.api_call('groups.info', data={'channel': self.id})["group"]
        else:
            return self._bot.api_call('channels.info', data={'channel': self.id})["channel"]

    @property
    def private(self):
        """Return True if the room is a private group"""
        return self._channel.id.startswith('G')

    @property
    def id(self):
        """Return the ID of this room"""
        if self._id is None:
            self._id = self._channel.id
        return self._id

    channelid = id

    @property
    def name(self):
        """Return the name of this room"""
        return self._name

    def join(self, username=None, password=None):
        log.info("Joining channel %s", str(self))
        try:
            self._bot.api_call('channels.join', data={'name': self.name})
        except SlackAPIResponseError as e:
            if e.error == 'user_is_bot':
                raise RoomError(f'Unable to join channel. {USER_IS_BOT_HELPTEXT}')
            else:
                raise RoomError(e)

    def leave(self, reason=None):
        try:
            if self.id.startswith('C'):
                log.info('Leaving channel %s (%s)', self, self.id)
                self._bot.api_call('channels.leave', data={'channel': self.id})
            else:
                log.info('Leaving group %s (%s)', self, self.id)
                self._bot.api_call('groups.leave', data={'channel': self.id})
        except SlackAPIResponseError as e:
            if e.error == 'user_is_bot':
                raise RoomError(f'Unable to leave channel. {USER_IS_BOT_HELPTEXT}')
            else:
                raise RoomError(e)
        self._id = None

    def create(self, private=False):
        try:
            if private:
                log.info('Creating group %s.', self)
                self._bot.api_call('groups.create', data={'name': self.name})
            else:
                log.info('Creating channel %s.', self)
                self._bot.api_call('channels.create', data={'name': self.name})
        except SlackAPIResponseError as e:
            if e.error == 'user_is_bot':
                raise RoomError(f"Unable to create channel. {USER_IS_BOT_HELPTEXT}")
            else:
                raise RoomError(e)

    def destroy(self):
        try:
            if self.id.startswith('C'):
                log.info('Archiving channel %s (%s)', self, self.id)
                self._bot.api_call('channels.archive', data={'channel': self.id})
            else:
                log.info('Archiving group %s (%s)', self, self.id)
                self._bot.api_call('groups.archive', data={'channel': self.id})
        except SlackAPIResponseError as e:
            if e.error == 'user_is_bot':
                raise RoomError(f'Unable to archive channel. {USER_IS_BOT_HELPTEXT}')
            else:
                raise RoomError(e)
        self._id = None

    @property
    def exists(self):
        channels = self._bot.channels(joined_only=False, exclude_archived=False)
        return len([c for c in channels if c['name'] == self.name]) > 0

    @property
    def joined(self):
        channels = self._bot.channels(joined_only=True)
        return len([c for c in channels if c['name'] == self.name]) > 0

    @property
    def topic(self):
        if self._channel_info['topic']['value'] == '':
            return None
        else:
            return self._channel_info['topic']['value']

    @topic.setter
    def topic(self, topic):
        if self.private:
            log.info('Setting topic of %s (%s) to %s.', self, self.id, topic)
            self._bot.api_call('groups.setTopic', data={'channel': self.id, 'topic': topic})
        else:
            log.info('Setting topic of %s (%s) to %s.', self, self.id, topic)
            self._bot.api_call('channels.setTopic', data={'channel': self.id, 'topic': topic})

    @property
    def purpose(self):
        if self._channel_info['purpose']['value'] == '':
            return None
        else:
            return self._channel_info['purpose']['value']

    @purpose.setter
    def purpose(self, purpose):
        if self.private:
            log.info('Setting purpose of %s (%s) to %s.', self, self.id, purpose)
            self._bot.api_call('groups.setPurpose', data={'channel': self.id, 'purpose': purpose})
        else:
            log.info('Setting purpose of %s (%s) to %s.', str(self), self.id, purpose)
            self._bot.api_call('channels.setPurpose', data={'channel': self.id, 'purpose': purpose})

    @property
    def occupants(self):
        members = self._channel_info['members']
        return [SlackRoomOccupant(self.sc, m, self.id, self._bot) for m in members]

    def invite(self, *args):
        users = {user['name']: user['id'] for user in self._bot.api_call('users.list')['members']}
        for user in args:
            if user not in users:
                raise UserDoesNotExistError(f'User "{user}" not found.')
            log.info('Inviting %s into %s (%s)', user, self, self.id)
            method = 'groups.invite' if self.private else 'channels.invite'
            response = self._bot.api_call(
                method,
                data={'channel': self.id, 'user': users[user]},
                raise_errors=False
            )

            if not response['ok']:
                if response['error'] == 'user_is_bot':
                    raise RoomError(f'Unable to invite people. {USER_IS_BOT_HELPTEXT}')
                elif response['error'] != 'already_in_channel':
                    raise SlackAPIResponseError(error=f'Slack API call to {method} failed: {response["error"]}.')

    def __eq__(self, other):
        if not isinstance(other, SlackRoom):
            return False
        return self.id == other.id



from errbot import BotPlugin, botcmd


class Example(BotPlugin):
    """
    This is a very basic plugin to try out your new installation and get you started.
    Feel free to tweak me to experiment with Errbot.
    You can find me in your init directory in the subdirectory plugins.
    """

    @botcmd  # flags a command
    def tryme(self, msg, args):  # a command callable with !tryme
        """
        Execute to check if Errbot responds to command.
        Feel free to tweak me to experiment with Errbot.
        You can find me in your init directory in the subdirectory plugins.
        """
        return 'It *works* !'  # This string format is markdown.



from errbot.storage import StoreMixin
from errbot.storage.memory import MemoryStoragePlugin


def test_simple_store_retreive():
    sm = StoreMixin()
    sm.open_storage(MemoryStoragePlugin(None), 'ns')
    sm['toto'] = 'titui'
    assert sm['toto'] == 'titui'


def test_mutable():
    sm = StoreMixin()
    sm.open_storage(MemoryStoragePlugin(None), 'ns')
    sm['toto'] = [1, 3]

    with sm.mutable('toto') as titi:
        titi[1] = 5

    assert sm['toto'] == [1, 5]



import textwrap
import subprocess

from dulwich import errors as dulwich_errors
from errbot import BotPlugin, botcmd
from errbot.utils import git_tag_list
from errbot.version import VERSION


class Help(BotPlugin):
    MSG_HELP_TAIL = 'Type help <command name> to get more info ' \
                    'about that specific command.'
    MSG_HELP_UNDEFINED_COMMAND = 'That command is not defined.'

    def is_git_directory(self, path='.'):
        try:
            tags = git_tag_list(path)
        except dulwich_errors.NotGitRepository:
            tags = None
        except Exception as _:
            # we might want to handle other exceptions another way. For now leaving this general
            tags = None

        return tags.pop(-1) if tags is not None else None

    # noinspection PyUnusedLocal
    @botcmd(template='about')
    def about(self, msg, args):
        """Return information about this Errbot instance and version"""
        git_version = self.is_git_directory()
        if git_version:
            return dict(version=f"{git_version.decode('utf-8')} GIT CHECKOUT")
        else:
            return {'version': VERSION}

    # noinspection PyUnusedLocal
    @botcmd
    def apropos(self, msg, args):
        """   Returns a help string listing available options.

        Automatically assigned to the "help" command."""
        if not args:
            return 'Usage: ' + self._bot.prefix + 'apropos search_term'

        description = 'Available commands:\n'

        cls_commands = {}
        for (name, command) in self._bot.all_commands.items():
            cls = self._bot.get_plugin_class_from_method(command)
            cls = str.__module__ + '.' + cls.__name__  # makes the fuul qualified name
            commands = cls_commands.get(cls, [])
            if not self.bot_config.HIDE_RESTRICTED_COMMANDS or self._bot.check_command_access(msg, name)[0]:
                commands.append((name, command))
                cls_commands[cls] = commands

        usage = ''
        for cls in sorted(cls_commands):
            commands = []
            for (name, command) in cls_commands[cls]:
                if name == 'help':
                    continue

                if command._err_command_hidden:
                    continue

                doc = command.__doc__

                if doc is not None and args.lower() not in doc.lower():
                    continue

                name_with_spaces = name.replace('_', ' ', 1)
                doc = (doc or '(undocumented)').strip().split('\n', 1)[0]
                commands.append('\t' + self._bot.prefix + name_with_spaces + ': ' + doc)

            usage += '\n'.join(commands)
        usage += '\n\n'

        return ''.join(filter(None, [description, usage])).strip()

    @botcmd
    def help(self, msg, args):
        """Returns a help string listing available options.
        Automatically assigned to the "help" command."""

        def may_access_command(m, cmd):
            m, _, _ = self._bot._process_command_filters(
                msg=m,
                cmd=cmd,
                args=None,
                dry_run=True
            )
            return m is not None

        def get_name(named):
            return named.__name__.lower()

        # Normalize args to lowercase for ease of use
        args = args.lower() if args else ''
        usage = ''
        description = '### All commands\n'

        cls_obj_commands = {}
        for (name, command) in self._bot.all_commands.items():
            cls = self._bot.get_plugin_class_from_method(command)
            obj = command.__self__
            _, commands = cls_obj_commands.get(cls, (None, []))
            if not self.bot_config.HIDE_RESTRICTED_COMMANDS or may_access_command(msg, name):
                commands.append((name, command))
                cls_obj_commands[cls] = (obj, commands)

        # show all
        if not args:
            for cls in sorted(cls_obj_commands.keys(), key=lambda c: cls_obj_commands[c][0].name):
                obj, commands = cls_obj_commands[cls]
                name = obj.name
                # shows class and description
                usage += f'\n**{name}**\n\n*{cls.__errdoc__.strip() or ""}*\n\n'

                for name, command in sorted(commands):
                    if command._err_command_hidden:
                        continue
                    # show individual commands
                    usage += self._cmd_help_line(name, command)
            usage += '\n\n'  # end cls section
        elif args:
            for cls, (obj, cmds) in cls_obj_commands.items():
                if obj.name.lower() == args:
                    break
            else:
                cls, obj, cmds = None, None, None

            if cls is None:
                # Plugin not found.
                description = ''
                all_commands = dict(self._bot.all_commands)
                all_commands.update(
                    {k.replace('_', ' '): v for k, v in all_commands.items()})
                if args in all_commands:
                    usage += self._cmd_help_line(args, all_commands[args], True)
                else:
                    usage += self.MSG_HELP_UNDEFINED_COMMAND
            else:
                # filter out the commands related to this class
                description = f'\n**{obj.name}**\n\n*{cls.__errdoc__.strip() or ""}*\n\n'
                pairs = []
                for (name, command) in cmds:
                    if self.bot_config.HIDE_RESTRICTED_COMMANDS:
                        if command._err_command_hidden:
                            continue

                        if not may_access_command(msg, name):
                            continue
                    pairs.append((name, command))

                pairs = sorted(pairs)

                for (name, command) in pairs:
                    usage += self._cmd_help_line(name, command)

        return ''.join(filter(None, [description, usage]))

    def _cmd_help_line(self, name, command, show_doc=False):
        """
        Returns:
            str. a single line indicating the help representation of a command.
        """
        cmd_name = name.replace('_', ' ')
        cmd_doc = textwrap.dedent(self._bot.get_doc(command)).strip()
        prefix = self._bot.prefix if getattr(command, '_err_command_prefix_required', True) else ''

        name = cmd_name
        patt = getattr(command, '_err_command_re_pattern', None)

        if patt:
            re_help_name = getattr(command, '_err_command_re_name_help', None)
            name = re_help_name if re_help_name else patt.pattern

        if not show_doc:
            cmd_doc = cmd_doc.split('\n')[0]

            if len(cmd_doc) > 80:
                cmd_doc = f'{cmd_doc[:77]}...'

        help_str = f'- **{prefix}{name}** - {cmd_doc}\n'

        return help_str



from os import path

extra_plugin_dir = path.join(path.dirname(path.realpath(__file__)), 'syntax_plugin')


def test_nosyntax(testbot):
    assert testbot.bot.commands['foo_nosyntax']._err_command_syntax is None


def test_syntax(testbot):
    assert testbot.bot.commands['foo']._err_command_syntax == '[optional] <mandatory>'


def test_re_syntax(testbot):
    assert testbot.bot.re_commands['re_foo']._err_command_syntax == '.*'


def test_arg_syntax(testbot):
    assert testbot.bot.commands['arg_foo']._err_command_syntax == '[-h] [--repeat-count REPEAT] value'



from pygments.style import Style
from pygments.token import Keyword, Name, Comment, String, \
     Number, Operator, Generic, Whitespace


class ErrStyle(Style):
    """
    A Pygments style based on the "friendly" theme
    """

    background_color = "#ffffcc"
    default_style = ""

    styles = {
        Whitespace:                "#3e4349",
        Comment:                   "#3f6b5b",
        # Comment.Preproc:           "noitalic #007020",
        # Comment.Special:           "noitalic bg:#fff0f0",

        Keyword:                   "bold #f06f00",
        # Keyword.Pseudo:            "nobold",
        # Keyword.Type:              "nobold #902000",

        Operator:                  "#3e4349",
        # Operator.Word:             "bold #007020",

        Name:                      "#3e4349",
        Name.Builtin:              "#007020",
        Name.Function:             "bold #3e4349",
        Name.Class:                "bold #3e4349",
        # Name.Namespace:            "bold #f07e2a",
        # Name.Exception:            "#007020",
        Name.Variable:             "underline #8a2be2",
        Name.Constant:             "underline #b91f49",
        # Name.Label:                "bold #002070",
        Name.Entity:               "bold #330000",
        # Name.Attribute:            "#4070a0",
        Name.Tag:                  "bold #f06f00",
        Name.Decorator:            "bold italic #3e4349",

        # String:                    "#3e4349",
        String:                    "#9a5151",
        # String.Doc:                "italic #3f65b5",
        String.Doc:                "italic #3f6b5b",
        # String.Doc:                "italic #9a7851",
        # String.Doc:                "italic #9a5151",
        # String.Interpol:           "italic #70a0d0",
        # String.Escape:             "bold #4070a0",
        # String.Regex:              "#235388",
        # String.Symbol:             "#517918",
        # String.Other:              "#c65d09",
        Number:                    "underline #9a5151",

        Generic:                   "#3e4349",
        Generic.Heading:           "bold #1014ad",
        Generic.Subheading:        "bold #1014ad",
        Generic.Deleted:           "bg:#c8f2ea #2020ff",
        Generic.Inserted:          "#3e4349",
        # Generic.Error:             "#FF0000",
        # Generic.Emph:              "italic",
        # Generic.Strong:            "bold",
        # Generic.Prompt:            "bold #c65d09",
        # Generic.Output:            "#888",
        # Generic.Traceback:         "#04D",

        # Error:                     "border:#FF0000"
    }



#!/usr/bin/env python3
import configparser
import json
import logging
import os
import pathlib
import sys
import time
from datetime import datetime

import requests
from requests.auth import HTTPBasicAuth

logging.basicConfig()

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)

DEFAULT_AVATAR = 'https://upload.wikimedia.org/wikipedia/commons/5/5f/Err-logo.png'

user_cache = {}

try:
    with open('user_cache', 'r') as f:
        user_cache = eval(f.read())
except FileNotFoundError:
    # File doesn't exist, so we continue on
    log.info("No user cache existing, will be generating it for the first time.")


def get_auth():
    """Get auth creds from Github Token

    token is generated from the personal tokens in github
    """
    token_file = pathlib.Path('token')
    token_env = os.getenv('ERRBOT_REPOS_TOKEN')

    if token_file.is_file():
        try:
            token_info = open('token', 'r').read()
        except ValueError:
            log.fatal("Token file cannot be properly read, should be of the form username:token")
            sys.exit(-1)
    elif token_env:
        token_info = token_env
    else:
        msg = "No 'token' file or environment variable 'ERROBOT_REPOS_TOKEN' found."
        log.fatal(msg)
        sys.exit(-1)

    try:
        user, token = token_info.strip().split(':')
    except ValueError:
        msg = "Token file cannot be properly read, should be of the form username:token"
        log.fatal(msg)
        sys.exit(-1)

    auth = HTTPBasicAuth(user, token)
    return auth


AUTH = get_auth()


def add_blacklisted(repo):
    with open('blacklisted.txt', 'a') as f:
        f.write(repo)
        f.write('\n')


plugins = {}


def save_plugins():
    with open('repos.json', 'w') as f:
        json.dump(plugins, f, indent=2, separators=(',', ': '))


BLACKLISTED = []
try:
    with open('blacklisted.txt', 'r') as f:
        BLACKLISTED = [line.strip() for line in f.readlines()]
except FileNotFoundError:
    log.info("No blacklisted.txt found, no plugins will be blacklisted.")


def get_avatar_url(repo):
    username = repo.split('/')[0]
    if username in user_cache:
        user = user_cache[username]
    else:
        user_res = requests.get('https://api.github.com/users/' + username, auth=AUTH)
        user = user_res.json()
        if 'avatar_url' in user:  # don't pollute the presistent cache
            user_cache[username] = user
            with open('user_cache', 'w') as f:
                f.write(repr(user_cache))
        rate_limit(user_res)
    return user['avatar_url'] if 'avatar_url' in user else DEFAULT_AVATAR


def rate_limit(resp):
    """
    Wait enough to be in the budget for this request.
    :param resp: the http response from github
    :return:
    """
    if 'X-RateLimit-Remaining' not in resp.headers:
        log.info("No rate limit detected. Hum along...")
        return
    remain = int(resp.headers['X-RateLimit-Remaining'])
    limit = int(resp.headers['X-RateLimit-Limit'])
    log.info('Rate limiter: %s allowed out of %d', remain, limit)
    if remain > 1:  # margin by one request
        return
    reset = int(resp.headers['X-RateLimit-Reset'])
    ts = datetime.fromtimestamp(reset)
    delay = (ts - datetime.now()).total_seconds()
    log.info("Hit rate limit. Have to wait for %d seconds", delay)
    if delay < 0:  # time drift
        delay = 2
    time.sleep(delay)


def parse_date(gh_date: str) -> datetime:
    return datetime.strptime(gh_date, "%Y-%m-%dT%H:%M:%SZ")


def check_repo(repo):
    repo_name = repo.get('full_name', None)
    if repo_name is None:
        log.error('No name in %s', repo)
    log.debug('Checking %s...', repo_name)
    code_resp = requests.get('https://api.github.com/search/code?q=extension:plug+repo:%s' % repo_name, auth=AUTH)
    if code_resp.status_code != 200:
        log.error('Error getting https://api.github.com/search/code?q=extension:plug+repo:%s', repo_name)
        log.error('code %d', code_resp.status_code)
        log.error('content %s', code_resp.text)

        return
    plug_items = code_resp.json()['items']
    if not plug_items:
        log.debug('No plugin found in %s, blacklisting it.', repo_name)
        add_blacklisted(repo_name)
        return
    owner = repo['owner']
    avatar_url = owner['avatar_url'] if 'avatar_url' in owner else DEFAULT_AVATAR

    days_old = (datetime.now() - parse_date(repo['updated_at'])).days
    score = repo['stargazers_count'] + repo['watchers_count'] * 2 + repo['forks_count'] - days_old / 25

    for plug in plug_items:
        plugfile_resp = requests.get('https://raw.githubusercontent.com/%s/master/%s' % (repo_name, plug['path']))
        log.debug('Found a plugin:')
        log.debug('Repo:  %s', repo_name)
        log.debug('File:  %s', plug['path'])
        parser = configparser.ConfigParser()
        try:
            parser.read_string(plugfile_resp.text)

            name = parser['Core']['Name']
            log.debug('Name: %s', name)

            if 'Documentation' in parser and 'Description' in parser['Documentation']:
                doc = parser['Documentation']['Description']
                log.debug('Documentation: %s', doc)
            else:
                doc = ''

            if 'Python' in parser:
                python = parser['Python']['Version']
                log.debug('Python Version: %s', python)
            else:
                python = '2'

            plugin = {
                'path': plug['path'],
                'repo': repo['html_url'],
                'documentation': doc,
                'name': name,
                'python': python,
                'avatar_url': avatar_url,
                'score': score,
            }

            repo_entry = plugins.get(repo_name, {})
            repo_entry[name] = plugin
            plugins[repo_name] = repo_entry
            log.debug('Catalog added plugin %s.', plugin['name'])
        except:
            log.error('Invalid syntax in %s, skipping... ' % plug['path'])
            continue

        rate_limit(plugfile_resp)

    save_plugins()
    rate_limit(code_resp)


def find_plugins(query):
    url = 'https://api.github.com/search/repositories?q=%s+in:name+language:python&sort=stars&order=desc' % query
    while True:
        repo_resp = requests.get(url, auth=AUTH)
        repo_json = repo_resp.json()
        if repo_json.get('message', None) == 'Bad credentials':
            log.error('Invalid credentials, check your token file, see README.')
            sys.exit(-1)
        log.debug("Repo reqs before ratelimit %s/%s" % (repo_resp.headers['X-RateLimit-Remaining'],
                                                        repo_resp.headers['X-RateLimit-Limit']))
        if 'message' in repo_json and repo_json['message'].startswith('API rate limit exceeded for'):
            log.error('API rate limit hit anyway ... wait for 30s')
            time.sleep(30)
            continue
        items = repo_json['items']

        for repo in items:
            if repo['full_name'] in BLACKLISTED:
                log.debug('Skipping %s.', repo)
                continue
            check_repo(repo)
        if 'next' not in repo_resp.links:
            break
        url = repo_resp.links['next']['url']
        log.debug('Next url: %s', url)
        rate_limit(repo_resp)


def main():
    find_plugins('err')
    # Those are found by global search only available on github UI:
    # https://github.com/search?l=&q=Documentation+extension%3Aplug&ref=advsearch&type=Code&utf8=%E2%9C%93
    url = 'https://api.github.com/repos/%s'
    with open('extras.txt', 'r') as extras:
        for repo_name in extras:
            repo_name = repo_name.strip()
            repo_resp = requests.get(url % repo_name, auth=AUTH)
            repo = repo_resp.json()
            if repo.get('message', None) == 'Bad credentials':
                log.error('Invalid credentials, check your token file, see README.')
                sys.exit(-1)
            if 'message' in repo and repo['message'].startswith('API rate limit exceeded for'):
                log.error('API rate limit hit anyway ... wait for 30s')
                time.sleep(30)
                continue
            if 'message' in repo and repo['message'].startswith('Not Found'):
                log.error('%s not found.', repo_name)
            else:
                check_repo(repo)
            rate_limit(repo_resp)


if __name__ == "__main__":
    main()



from os import path
import time

CURRENT_FILE_DIR = path.dirname(path.realpath(__file__))
extra_plugin_dir = path.join(CURRENT_FILE_DIR, 'poller_plugin')


def test_delayed_hello(testbot):
    assert 'Hello, world!' in testbot.exec_command('!hello')
    time.sleep(1)
    delayed_msg = 'Hello world! was sent 5 seconds ago'
    assert delayed_msg in testbot.pop_message(timeout=1)
    # Assert that only one message has been enqueued
    assert testbot.bot.outgoing_message_queue.empty()


def test_delayed_hello_loop(testbot):
    assert 'Hello, world!' in testbot.exec_command('!hello_loop')
    time.sleep(1)
    delayed_msg = 'Hello world! was sent 5 seconds ago'
    assert delayed_msg in testbot.pop_message(timeout=1)
    # Assert that only one message has been enqueued
    assert testbot.bot.outgoing_message_queue.empty()



import logging
import sys
from time import sleep
import re

import copyreg
from ansi.color import fg, fx
from pygments import highlight
from pygments.formatters import Terminal256Formatter
from pygments.lexers import get_lexer_by_name

from errbot.rendering import ansi, text, xhtml, imtext
from errbot.rendering.ansiext import enable_format, ANSI_CHRS, AnsiExtension
from errbot.backends.base import Message, Person, Presence, ONLINE, OFFLINE, Room, RoomOccupant
from errbot.core import ErrBot
from errbot.logs import console_hdlr

from markdown import Markdown
from markdown.extensions.extra import ExtraExtension

log = logging.getLogger(__name__)

ENCODING_INPUT = sys.stdin.encoding
ANSI = hasattr(sys.stderr, 'isatty') and sys.stderr.isatty()

enable_format('borderless', ANSI_CHRS, borders=False)


def borderless_ansi():
    """This makes a converter from markdown to ansi (console) format.
    It can be called like this:
    from errbot.rendering import ansi
    md_converter = ansi()  # you need to cache the converter

    ansi_txt = md_converter.convert(md_txt)
    """
    md = Markdown(output_format='borderless', extensions=[ExtraExtension(), AnsiExtension()])
    md.stripTopLevelTags = False
    return md


class TextPerson(Person):
    """
    Simple Person implementation which represents users as simple text strings.
    """

    def __init__(self, person, client=None, nick=None, fullname=None):
        self._person = person
        self._client = client
        self._nick = nick
        self._fullname = fullname

    @property
    def person(self):
        return self._person

    @property
    def client(self):
        return self._client

    @property
    def nick(self):
        return self._nick

    @property
    def fullname(self):
        return self._fullname

    @property
    def aclattr(self):
        return str(self)

    def __str__(self):
        return '@' + self._person

    def __eq__(self, other):
        if not isinstance(other, Person):
            return False
        return self.person == other.person

    def __hash__(self):
        return self.person.__hash__()


class TextRoom(Room):

    def __init__(self, name, bot):
        self._topic = ''
        self._joined = False
        self.name = name
        self._bot = bot

        # fill up the room with a coherent set of identities.
        self._occupants = [TextOccupant('somebody', self),
                           TextOccupant(TextPerson(bot.bot_config.BOT_ADMINS[0]), self),
                           TextOccupant(bot.bot_identifier, self)]

    def join(self, username=None, password=None):
        self._joined = True

    def leave(self, reason=None):
        self._joined = False

    def create(self):
        self._joined = True

    def destroy(self):
        self._joined = False

    @property
    def exists(self):
        return True

    @property
    def joined(self):
        return self._joined

    @property
    def topic(self):
        return self._topic

    @topic.setter
    def topic(self, topic):
        self._topic = topic

    @property
    def occupants(self):
        return self._occupants

    def invite(self, *args):
        pass

    def __str__(self):
        return '#' + self.name

    def __eq__(self, other):
        return self.name == other.name

    def __hash__(self):
        return self.name.__hash__()


class TextOccupant(TextPerson, RoomOccupant):

    def __init__(self, person, room):
        super().__init__(person)
        self._room = room

    @property
    def room(self):
        return self._room

    def __str__(self):
        return f'#{self._room.name}/{self._person.person}'

    def __eq__(self, other):
        return self.person == other.person and self.room == other.room

    def __hash__(self):
        return self.person.__hash__() + self.room.__hash__()


INTRO = """
---
You start as a **bot admin in a one-on-one conversation** with the bot.

### Context of the chat

- Use `!inroom`{:color='blue'} to switch to a room conversation.
- Use `!inperson`{:color='blue'} to switch back to a one-on-one conversation.
- Use `!asuser`{:color='green'} to talk as a normal user.
- Use `!asadmin`{:color='red'} to switch back as a bot admin.

### Preferences

- Use `!ml`{:color='yellow'} to flip on/off the multiline mode (Enter twice at the end to send).

---
"""


class TextBackend(ErrBot):
    def __init__(self, config):
        super().__init__(config)
        log.debug("Text Backend Init.")

        if hasattr(self.bot_config, 'BOT_IDENTITY') and 'username' in self.bot_config.BOT_IDENTITY:
            self.bot_identifier = self.build_identifier(self.bot_config.BOT_IDENTITY['username'])
        else:
            # Just a default identity for the bot if nothing has been specified.
            self.bot_identifier = self.build_identifier('@errbot')

        log.debug('Bot username set at %s.', self.bot_identifier)
        self._inroom = False
        self._rooms = []
        self._multiline = False

        self.demo_mode = self.bot_config.TEXT_DEMO_MODE if hasattr(self.bot_config, 'TEXT_DEMO_MODE') else False
        if not self.demo_mode:
            self.md_html = xhtml()  # for more debug feedback on md
            self.md_text = text()  # for more debug feedback on md
            self.md_borderless_ansi = borderless_ansi()
            self.md_im = imtext()
            self.md_lexer = get_lexer_by_name("md", stripall=True)

        self.md_ansi = ansi()
        self.html_lexer = get_lexer_by_name("html", stripall=True)
        self.terminal_formatter = Terminal256Formatter(style='paraiso-dark')
        self.user = self.build_identifier(self.bot_config.BOT_ADMINS[0])
        self._register_identifiers_pickling()

    @staticmethod
    def _unpickle_identifier(identifier_str):
        return TextBackend.__build_identifier(identifier_str)

    @staticmethod
    def _pickle_identifier(identifier):
        return TextBackend._unpickle_identifier, (str(identifier),)

    def _register_identifiers_pickling(self):
        """
        Register identifiers pickling.
        """
        TextBackend.__build_identifier = self.build_identifier
        for cls in (TextPerson, TextOccupant, TextRoom):
            copyreg.pickle(cls, TextBackend._pickle_identifier, TextBackend._unpickle_identifier)

    def serve_forever(self):
        self.readline_support()

        if not self._rooms:
            # artificially join a room if None were specified.
            self.query_room('#testroom').join()

        if self.demo_mode:
            # disable the console logging once it is serving in demo mode.
            root = logging.getLogger()
            root.removeHandler(console_hdlr)
            root.addHandler(logging.NullHandler())
        self.connect_callback()  # notify that the connection occured
        self.callback_presence(Presence(identifier=self.user, status=ONLINE))

        self.send_message(Message(INTRO))

        try:
            while True:

                if self._inroom:
                    frm = TextOccupant(self.user, self.rooms[0])
                    to = self.rooms[0]
                else:
                    frm = self.user
                    to = self.bot_identifier

                print()
                full_msg = ''
                while True:
                    prompt = '[␍] ' if full_msg else '>>> '
                    if ANSI or self.demo_mode:
                        color = fg.red if self.user.person in self.bot_config.BOT_ADMINS[0] else fg.green
                        prompt = f'{color}[{frm} ➡ {to}] {fg.cyan}{prompt}{fx.reset}'
                        entry = input(prompt)
                    else:
                        entry = input(f'[{frm} ➡ {to}] {prompt}')

                    if not self._multiline:
                        full_msg = entry
                        break

                    if not entry:
                        break

                    full_msg += entry + '\n'

                msg = Message(full_msg)
                msg.frm = frm
                msg.to = to

                self.callback_message(msg)

                mentioned = [self.build_identifier(word) for word in re.findall(r'(?<=\s)@[\w]+', entry)]
                if mentioned:
                    self.callback_mention(msg, mentioned)

                sleep(.5)
        except EOFError:
            pass
        except KeyboardInterrupt:
            pass
        finally:
            # simulate some real presence
            self.callback_presence(Presence(identifier=self.user, status=OFFLINE))
            log.debug("Trigger disconnect callback")
            self.disconnect_callback()
            log.debug("Trigger shutdown")
            self.shutdown()

    def readline_support(self):
        try:
            # Load readline for better editing/history behaviour
            import readline

            # Implement a simple completer for commands
            def completer(txt, state):
                options = [i for i in self.all_commands if i.startswith(txt)] if txt else list(self.all_commands.keys())
                if state < len(options):
                    return options[state]

            readline.parse_and_bind("tab: complete")
            readline.set_completer(completer)

        except ImportError:
            # Readline is Unix-only
            log.debug("Python readline module is not available")

    def send_message(self, msg):
        if self.demo_mode:
            print(self.md_ansi.convert(msg.body))
        else:
            bar = '\n╌╌[{mode}]' + ('╌' * 60)
            super().send_message(msg)
            print(bar.format(mode='MD  '))
            if ANSI:
                print(highlight(msg.body, self.md_lexer, self.terminal_formatter))
            else:
                print(msg.body)
            print(bar.format(mode='HTML'))
            html = self.md_html.convert(msg.body)
            if ANSI:
                print(highlight(html, self.html_lexer, self.terminal_formatter))
            else:
                print(html)
            print(bar.format(mode='TEXT'))
            print(self.md_text.convert(msg.body))
            print(bar.format(mode='IM  '))
            print(self.md_im.convert(msg.body))
            if ANSI:
                print(bar.format(mode='ANSI'))
                print(self.md_ansi.convert(msg.body))
                print(bar.format(mode='BORDERLESS'))
                print(self.md_borderless_ansi.convert(msg.body))
            print('\n\n')

    def add_reaction(self, msg: Message, reaction: str) -> None:
        # this is like the Slack backend's add_reaction
        self._react('+', msg, reaction)

    def remove_reaction(self, msg: Message, reaction: str) -> None:
        self._react('-', msg, reaction)

    def _react(self, sign, msg, reaction):
        self.send(msg.frm, f'reaction {sign}:{reaction}:', in_reply_to=msg)

    def change_presence(self, status: str = ONLINE, message: str = '') -> None:
        log.debug("*** Changed presence to [%s] %s", status, message)

    def build_identifier(self, text_representation):
        if text_representation.startswith('#'):
            rem = text_representation[1:]
            if '/' in text_representation:
                room, person = rem.split('/')
                return TextOccupant(TextPerson(person), TextRoom(room, self))
            return self.query_room('#' + rem)
        if not text_representation.startswith('@'):
            raise ValueError('An identifier for the Text backend needs to start with # for a room or @ for a person.')
        return TextPerson(text_representation[1:])

    def build_reply(self, msg, text=None, private=False, threaded=False):
        response = self.build_message(text)
        response.frm = self.bot_identifier
        if private:
            response.to = msg.frm
        else:
            response.to = msg.frm.room if isinstance(msg.frm, RoomOccupant) else msg.frm
        return response

    @property
    def mode(self):
        return 'text'

    def query_room(self, room):
        if not room.startswith('#'):
            raise ValueError('A Room name must start by #.')
        text_room = TextRoom(room[1:], self)
        if text_room not in self._rooms:
            self._rooms.insert(0, text_room)
        else:
            self._rooms.insert(0, self._rooms.pop(self._rooms.index(text_room)))
        return text_room

    @property
    def rooms(self):
        return self._rooms

    def prefix_groupchat_reply(self, message, identifier):
        message.body = f'{identifier.person} {message.body}'



from errbot import BotPlugin
from queue import Queue
import logging
log = logging.getLogger(__name__)


class RoomTest(BotPlugin):

    def activate(self):
        super().activate()
        self.purge()

    def callback_room_joined(self, room):
        log.info("join")
        self.events.put("callback_room_joined {!s}".format(room))

    def callback_room_left(self, room):
        self.events.put("callback_room_left {!s}".format(room))

    def callback_room_topic(self, room):
        self.events.put("callback_room_topic {}".format(room.topic))

    def purge(self):
        log.info("purge")
        self.events = Queue()



from errbot.backends.test import TestPerson, TestOccupant


def test_identifier_eq():
    a = TestPerson("foo")
    b = TestPerson("foo")
    assert a == b


def test_identifier_ineq():
    a = TestPerson("foo")
    b = TestPerson("bar")
    assert not a == b
    assert a != b


def test_mucidentifier_eq():
    a = TestOccupant("foo", "room")
    b = TestOccupant("foo", "room")
    assert a == b


def test_mucidentifier_ineq1():
    a = TestOccupant("foo", "room")
    b = TestOccupant("bar", "room")
    assert not a == b
    assert a != b


def test_mucidentifier_ineq2():
    a = TestOccupant("foo", "room1")
    b = TestOccupant("foo", "room2")
    assert not a == b
    assert a != b



import logging

from errbot import BotPlugin, botcmd, SeparatorArgParser, ShlexArgParser
from errbot.backends.base import RoomNotJoinedError

log = logging.getLogger(__name__)


class ChatRoom(BotPlugin):

    connected = False

    def callback_connect(self):
        self.log.info('Connecting bot chatrooms')
        if not self.connected:
            self.connected = True
            for room in self.bot_config.CHATROOM_PRESENCE:
                self.log.debug('Try to join room %s', repr(room))
                try:
                    self._join_room(room)
                except Exception:
                    # Ensure failure to join a room doesn't crash the plugin
                    # as a whole.
                    self.log.exception(f'Joining room {repr(room)} failed')

    def _join_room(self, room):
        username = self.bot_config.CHATROOM_FN
        password = None
        if isinstance(room, (tuple, list)):
            room, password = room  # unpack
            self.log.info('Joining room %s with username %s and pass ***.', room, username)
        else:
            self.log.info('Joining room %s with username %s.', room, username)
        self.query_room(room).join(username=self.bot_config.CHATROOM_FN, password=password)

    def deactivate(self):
        self.connected = False
        super().deactivate()

    @botcmd(split_args_with=SeparatorArgParser())
    def room_create(self, message, args):
        """
        Create a chatroom.

        Usage:
        !room create <room>

        Examples (XMPP):
        !room create example-room@chat.server.tld

        Examples (IRC):
        !room create #example-room
        """
        if len(args) < 1:
            return "Please tell me which chatroom to create."
        room = self.query_room(args[0])
        room.create()
        return f'Created the room {room}.'

    @botcmd(split_args_with=ShlexArgParser())
    def room_join(self, message, args):
        """
        Join (creating it first if needed) a chatroom.

        Usage:
        !room join <room> [<password>]

        Examples (XMPP):
        !room join example-room@chat.server.tld
        !room join example-room@chat.server.tld super-secret-password

        Examples (IRC):
        !room join #example-room
        !room join #example-room super-secret-password
        !room join #example-room "password with spaces"
        """
        arglen = len(args)
        if arglen < 1:
            return "Please tell me which chatroom to join."
        args[0].strip()

        room_name, password = (args[0], None) if arglen == 1 else (args[0], args[1])
        room = self.query_room(room_name)
        if room is None:
            return f'Cannot find room {room_name}.'

        room.join(username=self.bot_config.CHATROOM_FN, password=password)
        return f'Joined the room {room_name}.'

    @botcmd(split_args_with=SeparatorArgParser())
    def room_leave(self, message, args):
        """
        Leave a chatroom.

        Usage:
        !room leave <room>

        Examples (XMPP):
        !room leave example-room@chat.server.tld

        Examples (IRC):
        !room leave #example-room
        """
        if len(args) < 1:
            return 'Please tell me which chatroom to leave.'
        self.query_room(args[0]).leave()
        return f'Left the room {args[0]}.'

    @botcmd(split_args_with=SeparatorArgParser())
    def room_destroy(self, message, args):
        """
        Destroy a chatroom.

        Usage:
        !room destroy <room>

        Examples (XMPP):
        !room destroy example-room@chat.server.tld

        Examples (IRC):
        !room destroy #example-room
        """
        if len(args) < 1:
            return "Please tell me which chatroom to destroy."
        self.query_room(args[0]).destroy()
        return f'Destroyed the room {args[0]}.'

    @botcmd(split_args_with=SeparatorArgParser())
    def room_invite(self, message, args):
        """
        Invite one or more people into a chatroom.

        Usage:
        !room invite <room> <identifier1> [<identifier2>, ..]

        Examples (XMPP):
        !room invite room@conference.server.tld bob@server.tld

        Examples (IRC):
        !room invite #example-room bob
        """
        if len(args) < 2:
            return 'Please tell me which person(s) to invite into which room.'
        self.query_room(args[0]).invite(*args[1:])
        return f'Invited {", ".join(args[1:])} into the room {args[0]}.'

    @botcmd
    def room_list(self, message, args):
        """
        List chatrooms the bot has joined.

        Usage:
        !room list

        Examples:
        !room list
        """
        rooms = [str(room) for room in self.rooms()]
        if len(rooms):
            rooms_str = '\n\t'.join(rooms)
            return f"I'm currently in these rooms:\n\t{rooms_str}"
        else:
            return "I'm not currently in any rooms."

    @botcmd(split_args_with=ShlexArgParser())
    def room_occupants(self, message, args):
        """
        List the occupants in a given chatroom.

        Usage:
        !room occupants <room 1> [<room 2> ..]

        Examples (XMPP):
        !room occupants room@conference.server.tld

        Examples (IRC):
        !room occupants #example-room #another-example-room
        """
        if len(args) < 1:
            yield "Please supply a room to list the occupants of."
            return
        for room in args:
            try:
                occupants = [o.person for o in self.query_room(room).occupants]
                occupants_str = "\n\t".join(occupants)
                yield f'Occupants in {room}:\n\t{occupants_str}.'
            except RoomNotJoinedError as e:
                yield f'Cannot list occupants in {room}: {e}.'

    @botcmd(split_args_with=ShlexArgParser())
    def room_topic(self, message, args):
        """
        Get or set the topic for a room.

        Usage:
        !room topic <room> [<new topic>]

        Examples (XMPP):
        !room topic example-room@chat.server.tld
        !room topic example-room@chat.server.tld "Err rocks!"

        Examples (IRC):
        !room topic #example-room
        !room topic #example-room "Err rocks!"
        """
        arglen = len(args)
        if arglen < 1:
            return "Please tell me which chatroom you want to know the topic of."

        if arglen == 1:
            try:
                topic = self.query_room(args[0]).topic
            except RoomNotJoinedError as e:
                return f'Cannot get the topic for {args[0]}: {e}.'
            if topic is None:
                return f'No topic is set for {args[0]}.'
            else:
                return f'Topic for {args[0]}: {topic}.'
        else:
            try:
                self.query_room(args[0]).topic = args[1]
            except RoomNotJoinedError as e:
                return f'Cannot set the topic for {args[0]}: {e}.'
            return f"Topic for {args[0]} set."

    def callback_message(self, msg):
        try:
            if msg.is_direct:
                username = msg.frm.person
                if username in self.bot_config.CHATROOM_RELAY:
                    self.log.debug('Message to relay from %s.', username)
                    body = msg.body
                    rooms = self.bot_config.CHATROOM_RELAY[username]
                    for roomstr in rooms:
                        self.send(self.query_room(roomstr), body)
            elif msg.is_group:
                fr = msg.frm
                chat_room = str(fr.room)
                if chat_room in self.bot_config.REVERSE_CHATROOM_RELAY:
                    users_to_relay_to = self.bot_config.REVERSE_CHATROOM_RELAY[chat_room]
                    self.log.debug('Message to relay to %s.', users_to_relay_to)
                    body = f'[{fr.person}] {msg.body}'
                    for user in users_to_relay_to:
                        self.send(user, body)
        except Exception as e:
            self.log.exception(f'crashed in callback_message {e}')



import sys

import inspect
from configparser import ConfigParser
from dataclasses import dataclass
from importlib._bootstrap import module_from_spec
from importlib._bootstrap_external import spec_from_file_location

from errbot.utils import version2tuple
from pathlib import Path
from typing import Tuple, List, Type
from configparser import Error as ConfigParserError

VersionType = Tuple[int, int, int]


@dataclass
class PluginInfo:
    name: str
    module: str
    doc: str
    core: bool
    python_version: VersionType
    errbot_minversion: VersionType
    errbot_maxversion: VersionType
    dependencies: List[str]
    location: Path = None

    @staticmethod
    def load(plugfile_path: Path) -> 'PluginInfo':
        with plugfile_path.open(encoding='utf-8') as plugfile:
            return PluginInfo.load_file(plugfile, plugfile_path)

    @staticmethod
    def load_file(plugfile, location: Path) -> 'PluginInfo':
        cp = ConfigParser()
        cp.read_file(plugfile)
        pi = PluginInfo.parse(cp)
        pi.location = location
        return pi

    @staticmethod
    def parse(config: ConfigParser) -> 'PluginInfo':
        """
        Throws ConfigParserError with a meaningful message if the ConfigParser doesn't contain the minimal
         information required.
        """
        name = config.get('Core', 'Name')
        module = config.get('Core', 'Module')
        core = config.get('Core', 'Core', fallback='false').lower() == 'true'
        doc = config.get('Documentation', 'Description', fallback=None)

        python_version = config.get('Python', 'Version', fallback=None)
        # Old format backward compatibility
        if python_version:
            if python_version in ('2+', '3'):
                python_version = (3, 0, 0)
            elif python_version == '2':
                python_version = (2, 0, 0)
            else:
                try:
                    python_version = tuple(version2tuple(python_version)[0:3])  # We can ignore the alpha/beta part.
                except ValueError as ve:
                    raise ConfigParserError(f'Invalid Python Version format: {python_version} ({ve})')

        min_version = config.get('Errbot', 'Min', fallback=None)
        max_version = config.get('Errbot', 'Max', fallback=None)
        try:
            if min_version:
                min_version = version2tuple(min_version)
        except ValueError as ve:
            raise ConfigParserError(f'Invalid Errbot min version format: {min_version} ({ve})')

        try:
            if max_version:
                max_version = version2tuple(max_version)
        except ValueError as ve:
            raise ConfigParserError(f'Invalid Errbot max version format: {max_version} ({ve})')
        depends_on = config.get('Core', 'DependsOn', fallback=None)
        deps = [name.strip() for name in depends_on.split(',')] if depends_on else []

        return PluginInfo(name, module, doc, core, python_version, min_version, max_version, deps)

    def load_plugin_classes(self, base_module_name: str, baseclass: Type):
        # load the module
        module_name = base_module_name + '.' + self.module
        spec = spec_from_file_location(module_name, self.location.parent / (self.module + '.py'))
        modu1e = module_from_spec(spec)
        spec.loader.exec_module(modu1e)
        sys.modules[module_name] = modu1e

        # introspect the modules to find plugin classes
        def is_plugin(member):
            return inspect.isclass(member) and issubclass(member, baseclass) and member != baseclass

        plugin_classes = inspect.getmembers(modu1e, is_plugin)
        return plugin_classes



# coding=utf-8
import os
import re
import logging
from os import path, mkdir
from queue import Empty
from shutil import rmtree
from tempfile import mkdtemp
from mock import MagicMock

import pytest
import tarfile

extra_plugin_dir = path.join(path.dirname(path.realpath(__file__)), 'dummy_plugin')


def test_root_help(testbot):
    assert 'All commands' in testbot.exec_command('!help')


def test_help(testbot):
    assert '!about' in testbot.exec_command('!help Help')
    assert 'That command is not defined.' in testbot.exec_command('!help beurk')

    # Ensure that help reports on re_commands.
    assert 'runs foo' in testbot.exec_command('!help foo')  # Part of Dummy
    assert 'runs re_foo' in testbot.exec_command('!help re_foo')  # Part of Dummy
    assert 'runs re_foo' in testbot.exec_command('!help re foo')  # Part of Dummy


def test_about(testbot):
    assert 'Errbot version' in testbot.exec_command('!about')


def test_uptime(testbot):
    assert 'I\'ve been up for' in testbot.exec_command('!uptime')


def test_status(testbot):
    assert 'Yes I am alive' in testbot.exec_command('!status')


def test_status_plugins(testbot):
    assert 'A = Activated, D = Deactivated' in testbot.exec_command('!status plugins')


def test_status_load(testbot):
    assert 'Load ' in testbot.exec_command('!status load')


def test_whoami(testbot):
    assert 'person' in testbot.exec_command('!whoami')
    assert 'gbin@localhost' in testbot.exec_command('!whoami')


def test_echo(testbot):
    assert 'foo' in testbot.exec_command('!echo foo')


def test_status_gc(testbot):
    assert 'GC 0->' in testbot.exec_command('!status gc')


def test_config_cycle(testbot):
    testbot.push_message('!plugin config Webserver')
    m = testbot.pop_message()
    assert 'Default configuration for this plugin (you can copy and paste this directly as a command)' in m
    assert 'Current configuration' not in m

    testbot.assertInCommand("!plugin config Webserver {'HOST': 'localhost', 'PORT': 3141, 'SSL':  None}",
                            'Plugin configuration done.')

    assert 'Current configuration' in testbot.exec_command('!plugin config Webserver')
    assert 'localhost' in testbot.exec_command('!plugin config Webserver')


def test_apropos(testbot):
    assert '!about: Return information about' in testbot.exec_command('!apropos about')


def test_logtail(testbot):
    assert 'DEBUG' in testbot.exec_command('!log tail')


def test_history(testbot):
    assert 'up' in testbot.exec_command('!uptime')
    assert 'uptime' in testbot.exec_command('!history')

    orig_sender = testbot.bot.sender
    # Pretend to be someone else. History should be empty
    testbot.bot.sender = testbot.bot.build_identifier('non_default_person')
    testbot.push_message('!history')
    with pytest.raises(Empty):
        testbot.pop_message(timeout=1)
    assert 'should be a separate history' in testbot.exec_command('!echo should be a separate history')
    assert 'should be a separate history' in testbot.exec_command('!history')
    testbot.bot.sender = orig_sender
    # Pretend to be the original person again. History should still contain uptime
    assert 'uptime' in testbot.exec_command('!history')


def test_plugin_cycle(testbot):

    plugins = [
        'errbotio/err-helloworld',
    ]

    for plugin in plugins:
        testbot.assertInCommand(
            '!repos install {0}'.format(plugin),
            'Installing {0}...'.format(plugin)
        ),
        assert 'A new plugin repository has been installed correctly from errbotio/err-helloworld' in \
               testbot.pop_message(timeout=60)
        assert 'Plugins reloaded' in testbot.pop_message()

        assert 'this command says hello' in testbot.exec_command('!help hello')
        assert 'Hello World !' in testbot.exec_command('!hello')

        testbot.push_message('!plugin reload HelloWorld')
        assert 'Plugin HelloWorld reloaded.' == testbot.pop_message()

        testbot.push_message('!hello')  # should still respond
        assert 'Hello World !' == testbot.pop_message()

        testbot.push_message('!plugin blacklist HelloWorld')
        assert 'Plugin HelloWorld is now blacklisted.' == testbot.pop_message()
        testbot.push_message('!plugin deactivate HelloWorld')
        assert 'HelloWorld is already deactivated.' == testbot.pop_message()

        testbot.push_message('!hello')  # should not respond
        assert 'Command "hello" not found' in testbot.pop_message()

        testbot.push_message('!plugin unblacklist HelloWorld')
        assert 'Plugin HelloWorld removed from blacklist.' == testbot.pop_message()
        testbot.push_message('!plugin activate HelloWorld')
        assert 'HelloWorld is already activated.' == testbot.pop_message()

        testbot.push_message('!hello')  # should respond back
        assert 'Hello World !' == testbot.pop_message()

        testbot.push_message('!repos uninstall errbotio/err-helloworld')
        assert 'Repo errbotio/err-helloworld removed.' == testbot.pop_message()

        testbot.push_message('!hello')  # should not respond
        assert 'Command "hello" not found' in testbot.pop_message()


def test_broken_plugin(testbot):

    borken_plugin_dir = path.join(path.dirname(path.realpath(__file__)), 'borken_plugin')
    try:
        tempd = mkdtemp()
        tgz = os.path.join(tempd, "borken.tar.gz")
        with tarfile.open(tgz, "w:gz") as tar:
            tar.add(borken_plugin_dir, arcname='borken')
        assert 'Installing' in testbot.exec_command('!repos install file://' + tgz, timeout=120)
        assert 'import borken  # fails' in testbot.pop_message()
        assert 'as it did not load correctly.' in testbot.pop_message()
        assert 'Plugins reloaded.' in testbot.pop_message()
    finally:
        rmtree(tempd)


def test_backup(testbot):
    bot = testbot.bot  # used while restoring
    bot.push_message('!repos install https://github.com/errbotio/err-helloworld.git')
    assert 'Installing' in testbot.pop_message()
    assert 'err-helloworld' in testbot.pop_message(timeout=60)
    assert 'reload' in testbot.pop_message()
    bot.push_message('!backup')
    msg = testbot.pop_message()
    assert 'has been written in' in msg
    filename = re.search(r'"(.*)"', msg).group(1)

    # At least the backup should mention the installed plugin
    assert 'errbotio/err-helloworld' in open(filename).read()

    # Now try to clean the bot and restore
    for p in testbot.bot.plugin_manager.get_all_active_plugins():
        p.close_storage()

    assert 'Plugin HelloWorld deactivated.' in testbot.exec_command('!plugin deactivate HelloWorld')

    plugins_dir = path.join(testbot.bot_config.BOT_DATA_DIR, 'plugins')
    bot.repo_manager['installed_repos'] = {}
    bot.plugin_manager['configs'] = {}
    rmtree(plugins_dir)
    mkdir(plugins_dir)

    from errbot.bootstrap import restore_bot_from_backup
    log = logging.getLogger(__name__)  # noqa
    restore_bot_from_backup(filename, bot=bot, log=log)

    assert 'Plugin HelloWorld activated.' in testbot.exec_command('!plugin activate HelloWorld')
    assert 'Hello World !' in testbot.exec_command('!hello')
    testbot.push_message('!repos uninstall errbotio/err-helloworld')


def test_encoding_preservation(testbot):
    testbot.push_message('!echo へようこそ')
    assert 'へようこそ' == testbot.pop_message()


def test_webserver_webhook_test(testbot):
    testbot.push_message("!plugin config Webserver {'HOST': 'localhost', 'PORT': 3141, 'SSL':  None}")
    assert 'Plugin configuration done.' in testbot.pop_message()
    testbot.assertInCommand("!webhook test /echo toto", 'Status code : 200')


def test_activate_reload_and_deactivate(testbot):
    for command in ('activate', 'reload', 'deactivate'):
        testbot.push_message("!plugin {}".format(command))
        m = testbot.pop_message()
        assert 'Please tell me which of the following plugins to' in m
        assert 'ChatRoom' in m

        testbot.push_message(f'!plugin {command} nosuchplugin')
        m = testbot.pop_message()
        assert "nosuchplugin isn't a valid plugin name. The current plugins are" in m
        assert 'ChatRoom' in m

    testbot.push_message('!plugin reload ChatRoom')
    assert 'Plugin ChatRoom reloaded.' == testbot.pop_message()

    testbot.push_message('!status plugins')
    assert 'A      │ ChatRoom' in testbot.pop_message()

    testbot.push_message('!plugin deactivate ChatRoom')
    assert 'Plugin ChatRoom deactivated.' == testbot.pop_message()

    testbot.push_message("!status plugins")
    assert 'D      │ ChatRoom' in testbot.pop_message()

    testbot.push_message('!plugin deactivate ChatRoom')
    assert 'ChatRoom is already deactivated.' in testbot.pop_message()

    testbot.push_message('!plugin activate ChatRoom')
    assert 'Plugin ChatRoom activated.' in testbot.pop_message()

    testbot.push_message('!status plugins')
    assert 'A      │ ChatRoom' in testbot.pop_message()

    testbot.push_message('!plugin activate ChatRoom')
    assert 'ChatRoom is already activated.' == testbot.pop_message()

    testbot.push_message('!plugin deactivate ChatRoom')
    assert 'Plugin ChatRoom deactivated.' == testbot.pop_message()
    testbot.push_message('!plugin reload ChatRoom')
    assert 'Warning: plugin ChatRoom is currently not activated. Use !plugin activate ChatRoom to activate it.' == \
           testbot.pop_message()
    assert 'Plugin ChatRoom reloaded.' == testbot.pop_message()

    testbot.push_message('!plugin blacklist ChatRoom')
    assert 'Plugin ChatRoom is now blacklisted.' == testbot.pop_message()

    testbot.push_message('!status plugins')
    assert 'B,D    │ ChatRoom' in testbot.pop_message()

    # Needed else configuration for this plugin gets saved which screws up
    # other tests
    testbot.push_message('!plugin unblacklist ChatRoom')
    testbot.pop_message()


def test_unblacklist_and_blacklist(testbot):
    testbot.push_message('!plugin unblacklist nosuchplugin')
    m = testbot.pop_message()
    assert "nosuchplugin isn't a valid plugin name. The current plugins are" in m
    assert 'ChatRoom' in m

    testbot.push_message('!plugin blacklist nosuchplugin')
    m = testbot.pop_message()
    assert "nosuchplugin isn't a valid plugin name. The current plugins are" in m
    assert 'ChatRoom' in m

    testbot.push_message('!plugin blacklist ChatRoom')
    assert 'Plugin ChatRoom is now blacklisted' in testbot.pop_message()

    testbot.push_message('!plugin blacklist ChatRoom')
    assert 'Plugin ChatRoom is already blacklisted.' == testbot.pop_message()

    testbot.push_message('!status plugins')
    assert 'B,D    │ ChatRoom' in testbot.pop_message()

    testbot.push_message('!plugin unblacklist ChatRoom')
    assert 'Plugin ChatRoom removed from blacklist.' == testbot.pop_message()

    testbot.push_message('!plugin unblacklist ChatRoom')
    assert 'Plugin ChatRoom is not blacklisted.' == testbot.pop_message()

    testbot.push_message('!status plugins')
    assert 'A      │ ChatRoom' in testbot.pop_message()


def test_optional_prefix(testbot):
    testbot.bot_config.BOT_PREFIX_OPTIONAL_ON_CHAT = False
    assert 'Yes I am alive' in testbot.exec_command('!status')

    testbot.bot_config.BOT_PREFIX_OPTIONAL_ON_CHAT = True
    assert 'Yes I am alive' in testbot.exec_command('!status')
    assert 'Yes I am alive' in testbot.exec_command('status')


def test_optional_prefix_re_cmd(testbot):
    testbot.bot_config.BOT_PREFIX_OPTIONAL_ON_CHAT = False
    assert 'bar' in testbot.exec_command('!plz dont match this')

    testbot.bot_config.BOT_PREFIX_OPTIONAL_ON_CHAT = True
    assert 'bar' in testbot.exec_command('!plz dont match this')
    assert 'bar' in testbot.exec_command('plz dont match this')


def test_simple_match(testbot):
    assert 'bar' in testbot.exec_command('match this')


def test_no_suggest_on_re_commands(testbot):
    testbot.push_message('!re_ba')
    # Don't suggest a regexp command.
    assert '!re bar' not in testbot.pop_message()


def test_callback_no_command(testbot):
    extra_plugin_dir = os.path.join(
        os.path.dirname(os.path.realpath(__file__)),
        'commandnotfound_plugin'
    )

    cmd = '!this_is_not_a_real_command_at_all'
    expected_str = "Command fell through: {}".format(cmd)

    testbot.exec_command('!plugin deactivate CommandNotFoundFilter')
    testbot.bot.plugin_manager._extra_plugin_dir = extra_plugin_dir
    testbot.bot.plugin_manager.update_plugin_places([])
    testbot.exec_command('!plugin activate TestCommandNotFoundFilter')
    assert expected_str == testbot.exec_command(cmd)


def test_subcommands(testbot):
    # test single subcommand (method is run_subcommands())
    cmd = '!run subcommands with these args'
    cmd_underscore = '!run_subcommands with these args'
    expected_args = 'with these args'
    assert expected_args == testbot.exec_command(cmd)
    assert expected_args == testbot.exec_command(cmd_underscore)

    # test multiple subcmomands (method is run_lots_of_subcommands())
    cmd = '!run lots of subcommands with these args'
    cmd_underscore = '!run_lots_of_subcommands with these args'
    assert expected_args == testbot.exec_command(cmd)
    assert expected_args == testbot.exec_command(cmd_underscore)


def test_command_not_found_with_space_in_bot_prefix(testbot):
    testbot.bot_config.BOT_PREFIX = '! '
    assert 'Command "blah" not found.' in testbot.exec_command('! blah')
    assert 'Command "blah" / "blah toto" not found.' in testbot.exec_command('! blah toto')


def test_mock_injection(testbot):
    helper_mock = MagicMock()
    helper_mock.return_value = 'foo'
    mock_dict = {'helper_method': helper_mock}
    testbot.inject_mocks('Dummy', mock_dict)
    assert 'foo' in testbot.exec_command('!baz')


def test_multiline_command(testbot):
    testbot.assertInCommand(
        '''
        !bar title
        first line of body
        second line of body
        ''',
        '!bar title\nfirst line of body\nsecond line of body',
        dedent=True
    )


def test_plugin_info_command(testbot):
    output = testbot.exec_command('!plugin info Help')
    assert 'name: Help' in output
    assert 'module: help' in output
    assert 'help.py' in output
    assert 'log level: NOTSET' in output
